import unittest
import torch
from torch import Tensor
import numpy as np
from lightwood.encoder.categorical.onehot import (
    OneHotEncoder,
)

from lightwood.helpers.constants import _UNCOMMON_WORD


class TestOnehot(unittest.TestCase):
    """
    Test the OHE vector encoder

    NOTE, this is sensitive to the mapping of enc.map. If your unit-tests are failing, check to see if the enc.map order matches what the encoded_data should be; this is based on the mapped-place. Using np.unique() inadvertently sorts the data neatly, hence allowing an intuitive place of which order each category is.

    """ # noqa

    def test_encode_and_decode_with_unknown_token(self):
        """
        Tests the case where `use_unknown` is True; this means that the OHE vector is N categories + 1, where the first index represents UNK token.

        Checks:
        (1) UNKS are handled by "1" in the first index
        (2) Nones or unrecognized categories are both handled
        (3) The decode/encode map order is the same
        """ # noqa
        data = ['category 1', 'category 3', 'category 4', None, 'category 3']
        test_data = ['CATEGORY 1', 'category 2', 'category 1', 'category 3', None]

        ytest = [
            _UNCOMMON_WORD,
            _UNCOMMON_WORD,
            'category 1',
            'category 3',
            _UNCOMMON_WORD,
        ]

        enc = OneHotEncoder(use_unknown=True)
        enc.prepare(data)

        # Check the encoded patterns correct
        encoded_data = enc.encode(data)
        self.assertTrue(
            (
                encoded_data
                == Tensor(
                    [
                        [0.0, 1.0, 0.0, 0.0],  # category 1
                        [0.0, 0.0, 1.0, 0.0],  # category 3
                        [0.0, 0.0, 0.0, 1.0],  # category 4
                        [1.0, 0.0, 0.0, 0.0],  # None
                        [0.0, 0.0, 1.0, 0.0],  # category 3
                    ]
                )
            ).all()
        )

        # Check the decoded patterns correct
        decoded_data = enc.decode(enc.encode(test_data))
        for i in range(len(ytest)):
            self.assertTrue(decoded_data[i] == ytest[i])

    def test_encode_and_decode_with_return_zeros(self):
        """
        Tests the case where `use_unknown` is False; this means that the OHE vector is N categories, where a vector of all 0s represents the unknown token

        Checks:
        (1) UNKS are handled by a 0 vector returned
        (2) Nones or unrecognized categories are both handled
        (3) The decode/encode map order is the same
        """ # noqa

        data = ['category 1', 'category 3', 'category 4', None]
        test_data = ['category 2', 'category 1', 'category 3', None]

        enc = OneHotEncoder(use_unknown=False)
        enc.prepare(data)
        encoded_data = enc.encode(data)
        self.assertTrue(
            (
                encoded_data
                == Tensor(
                    [
                        [1.0, 0.0, 0.0],  # category 1
                        [0.0, 1.0, 0.0],  # category 3
                        [0.0, 0.0, 1.0],  # category 4
                        [0.0, 0.0, 0.0],  # None
                    ]
                )
            ).all()
        )

        decoded_data = enc.decode(enc.encode(test_data))
        self.assertEqual(
            decoded_data,
            [
                _UNCOMMON_WORD,  # category 2 is not seen, thus encoded to zero, thus decoded to uncommon word
                'category 1',  # seen before
                'category 3',  # seen before
                _UNCOMMON_WORD,  # None is encoded to zero, thus decoded to uncommon word
            ],
        )

    def test_check_probs_with_unknown(self):
        """ Check probability calculation where `use_unknown=False` """ # noqa
        data = ['category 1', 'category 3', 'category 4', None, 'category 3']

        enc = OneHotEncoder(use_unknown=False)
        enc.prepare(data)

        # Make data to represent random weights that do not sum to 1
        torch.manual_seed(1)
        wt_vec = torch.rand(size=(len(data), len(enc.map)))

        _, probs, _ = enc.decode_probabilities(wt_vec)
        self.assertTrue(np.all([np.isclose(sum(i), 1) for i in probs]))

    def test_target_distro_scaled_to_1(self):
        """
        Check whether target distribution passed and handled properly
        """ # noqa
        data = ["apple", "apple", "banana", "apple", "apple", "orange"]

        # Scaled weights (sum to 1)
        tweights = {"apple": 4 / 6, "banana": 1 / 6, "orange": 1 / 6}

        enc = OneHotEncoder(use_unknown=False, is_target=True, target_weights=tweights)
        enc.prepare(data)

        # Get the ground-truth weights
        iweights = torch.ones(size=(len(tweights),))
        for key, value in tweights.items():  # accounts for order
            iweights[enc.map[key]] = value

        # Check inverse weights correct
        self.assertTrue(np.all(((enc.index_weights - iweights) == 0).tolist()))

    def test_target_distro_with_unk(self):
        """ Checks target distro with unknowns """
        data = ["apple", "banana", "banana", "apple", "apple", "orange"]

        # Scaled weights (sum to 1)
        tweights = {"apple": 3 / 6, "orange": 1 / 6, "banana": 2 / 6}

        enc = OneHotEncoder(is_target=True, target_weights=tweights)
        enc.prepare(data)

        # Get the ground-truth inverse weights
        iweights = torch.ones(size=(len(tweights) + 1,))
        for key, value in tweights.items():  # accounts for order
            iweights[enc.map[key]] = value

        iweights[0] = min(iweights[1:])  # assign the smallest weight

        # Check inverse weights correct
        self.assertTrue(np.all(((enc.index_weights - iweights) == 0).tolist()))

    def test_distro_nonzeroweights(self):
        """
        Tests if target wts do not sum to 1 properly handled.

        This handles cases where people may choose something like 1/class_size for weights.

        Index-weight order is sensitive to the mapping
        """ # noqa

        data = ["apple", "apple", "banana", "apple", "apple", "orange"]
        tweights = {"apple": 100, "orange": 5000, "banana": 4}

        enc = OneHotEncoder(use_unknown=False, is_target=True, target_weights=tweights)
        enc.prepare(data)

        # Get the ground-truth weights
        iweights = torch.ones(size=(len(tweights),))
        for key, value in tweights.items():  # accounts for order
            iweights[enc.map[key]] = value

        for key, value in tweights.items():  # accounts for order
            iweights[enc.map[key]] = value

        # Reorder the weights with regards to the map
        self.assertTrue(np.all(((enc.index_weights - iweights) == 0).tolist()))

    def test_distro_zero(self):
        """
        Tests edge cause where target weights have a 0 weight which is unacceptable for downstream processing (inverse weights will 1/0)
        """  # noqa
        data = ["apple", "apple", "banana", "apple", "apple", "orange"]

        # Arbitrary weights (ex: number of examples)
        tweights = {"apple": 100, "banana": 1, "orange": 0}

        enc = OneHotEncoder(use_unknown=False, is_target=True, target_weights=tweights)

        # Check if 0-weight class is rejected
        self.assertRaises(ValueError, enc.prepare, data)



import unittest
import pandas as pd
from sklearn.metrics import accuracy_score
from lightwood.api.high_level import ProblemDefinition, json_ai_from_problem
from lightwood.api.high_level import code_from_json_ai, predictor_from_code

from lightwood.mixer import QClassic


class TestBasic(unittest.TestCase):
    def test_0_predict_file_flow(self):
        if QClassic is None:
            print('Skipping this test since the system for the quantum are not installed')
            return

        df = pd.read_csv('tests/data/hdi.csv')[:500]
        target = 'Development Index'

        pdef = ProblemDefinition.from_dict({'target': target, 'time_aim': 80})
        json_ai = json_ai_from_problem(df, problem_definition=pdef)

        neural_args = json_ai.model['args']['submodels'][0]['args']
        neural_args["target_encoder"] = "$encoders[self.target]"
        neural_args["target"] = "$target"
        neural_args["dtype_dict"] = "$dtype_dict"
        neural_args["input_cols"] = "$input_cols"
        neural_args["timeseries_settings"] = "$problem_definition.timeseries_settings"
        neural_args["net"] = '"DefaultNet"'

        json_ai.model['args']['submodels'] = [{
            'module': 'lightwood.mixer.QClassic',
            'args': neural_args
        }]

        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)

        predictor.learn(df)
        predictions = predictor.predict(df)
        return

        # sanity checks
        self.assertTrue(accuracy_score(df[target], predictions['prediction']) > 0.5)
        self.assertTrue(all([0 <= p <= 1 for p in predictions['confidence']]))



"""
"""
import time
import torch
from torch.utils.data import DataLoader
import os
import pandas as pd
from lightwood.encoder.text.helpers.pretrained_helpers import TextEmbed
from lightwood.helpers.device import get_devices
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.api import dtype
from transformers import (
    DistilBertModel,
    DistilBertForSequenceClassification,
    DistilBertTokenizerFast,
    AdamW,
    get_linear_schedule_with_warmup,
)
from lightwood.helpers.general import is_none
from typing import Iterable


class PretrainedLangEncoder(BaseEncoder):
    is_trainable_encoder: bool = True

    """
    Creates a contextualized embedding to represent input text via the [CLS] token vector from DistilBERT (transformers). (Sanh et al. 2019 - https://arxiv.org/abs/1910.01108).

    In certain text tasks, this model can use a transformer to automatically fine-tune on a class of interest (providing there is a 2 column dataset, where the input column is text).

    """ # noqa

    def __init__(
        self,
        stop_after: float,
        is_target: bool = False,
        batch_size: int = 10,
        max_position_embeddings: int = None,
        frozen: bool = False,
        epochs: int = 1,
        output_type: str = None,
        embed_mode: bool = True,
    ):
        """
        :param is_target: Whether this encoder represents the target. NOT functional for text generation yet.
        :param batch_size: size of batch while fine-tuning
        :param max_position_embeddings: max sequence length of input text
        :param custom_train: If True, trains model on target procided
        :param frozen: If True, freezes transformer layers during training.
        :param epochs: number of epochs to train model with
        :param output_type: Data dtype of the target; if categorical/binary, the option to return logits is possible.
        :param embed_mode: If True, assumes the output of the encode() step is the CLS embedding (this can be trained or not). If False, returns the logits of the tuned task.
        """ # noqa
        super().__init__(is_target)

        self.output_type = output_type
        self.name = "distilbert text encoder"

        self._max_len = max_position_embeddings
        self._frozen = frozen
        self._batch_size = batch_size
        self._epochs = epochs

        # Model setup
        self._model = None
        self.model_type = None

        # TODO: Other LMs; Distilbert is a good balance of speed/performance
        self._classifier_model_class = DistilBertForSequenceClassification
        self._embeddings_model_class = DistilBertModel
        self._pretrained_model_name = "distilbert-base-uncased"
        self._tokenizer = DistilBertTokenizerFast.from_pretrained(self._pretrained_model_name)

        self.device, _ = get_devices()
        self.stop_after = stop_after

        self.embed_mode = embed_mode
        self.uses_target = True
        self.output_size = None

        if self.embed_mode:
            log.info("Embedding mode on. [CLS] embedding dim output of encode()")
        else:
            log.info("Embedding mode off. Logits are output of encode()")

    def prepare(
        self,
        train_priming_data: Iterable[str],
        dev_priming_data: Iterable[str],
        encoded_target_values: torch.Tensor,
    ):
        """
        Fine-tunes a transformer on the priming data.

        CURRENTLY WIP; train + dev are placeholders for a validation-based approach. 

        Train + Dev are concatenated together and a transformer is then fine tuned with weight-decay applied on the transformer parameters. The option to freeze the underlying transformer and only train a linear layer exists if `frozen=True`. This trains faster, with the exception that the performance is often lower than fine-tuning on internal benchmarks.

        :param train_priming_data: Text data in the train set
        :param dev_priming_data: Text data in the dev set (not currently supported; can be empty)
        :param encoded_target_values: Encoded target labels in Nrows x N_output_dimension
        """ # noqa
        if self.is_prepared:
            raise Exception("Encoder is already prepared.")

        os.environ['TOKENIZERS_PARALLELISM'] = 'true'

        # TODO -> we shouldn't be concatenating these together
        if len(dev_priming_data) > 0:
            priming_data = pd.concat([train_priming_data, dev_priming_data]).values
        else:
            priming_data = train_priming_data.tolist()

        # Replaces empty strings with ''
        priming_data = [x if x is not None else "" for x in priming_data]

        # If classification, then fine-tune
        if (self.output_type in (dtype.categorical, dtype.binary)):
            log.info("Training model.")

            # Prepare priming data into tokenized form + attention masks
            text = self._tokenizer(priming_data, truncation=True, padding=True)

            log.info("\tOutput trained is categorical")

            # Label encode the OHE/binary output for classification
            labels = encoded_target_values.argmax(dim=1)

            # Construct the model
            self._model = self._classifier_model_class.from_pretrained(
                self._pretrained_model_name,
                num_labels=len(encoded_target_values[0]),  # max classes to test
            ).to(self.device)

            # Construct the dataset for training
            xinp = TextEmbed(text, labels)
            dataset = DataLoader(xinp, batch_size=self._batch_size, shuffle=True)

            # Set max length of input string; affects input to the model
            if self._max_len is None:
                self._max_len = self._model.config.max_position_embeddings

            if self._frozen:
                log.info("\tFrozen Model + Training Classifier Layers")
                """
                Freeze the base transformer model and train
                a linear layer on top
                """
                # Freeze all the transformer parameters
                for param in self._model.base_model.parameters():
                    param.requires_grad = False

                optimizer_grouped_parameters = self._model.parameters()

            else:
                log.info("\tFine-tuning model")
                """
                Fine-tuning parameters with weight decay
                """
                no_decay = [
                    "bias",
                    "LayerNorm.weight",
                ]  # decay on all terms EXCLUDING bias/layernorms
                optimizer_grouped_parameters = [
                    {
                        "params": [
                            p
                            for n, p in self._model.named_parameters()
                            if not any(nd in n for nd in no_decay)
                        ],
                        "weight_decay": 0.01,
                    },
                    {
                        "params": [
                            p
                            for n, p in self._model.named_parameters()
                            if any(nd in n for nd in no_decay)
                        ],
                        "weight_decay": 0.0,
                    },
                ]

            optimizer = AdamW(optimizer_grouped_parameters, lr=1e-5)
            scheduler = get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=0,  # default value for GLUE
                num_training_steps=len(dataset) * self._epochs,
            )

            # Train model; declare optimizer earlier if desired.
            self._tune_model(
                dataset, optim=optimizer, scheduler=scheduler, n_epochs=self._epochs
            )

        else:
            log.info("Target is not classification; Embeddings Generator only")

            self.model_type = "embeddings_generator"
            self._model = self._embeddings_model_class.from_pretrained(
                self._pretrained_model_name
            ).to(self.device)

            # TODO: Not a great flag
            # Currently, if the task is not classification, you must have
            # an embedding generator only.
            if self.embed_mode is False:
                log.info("Embedding mode must be ON for non-classification targets.")
                self.embed_mode = True

        self.is_prepared = True
        encoded = self.encode(priming_data[0:1])
        self.output_size = len(encoded[0])

    def _tune_model(self, dataset, optim, scheduler, n_epochs=1):
        """
        Given a model, train for n_epochs.
        Specifically intended for tuning; it does NOT use loss/
        stopping criterion.

        model - torch.nn model;
        dataset - torch.DataLoader; dataset to train
        device - torch.device; cuda/cpu
        log - lightwood.logger.log; log.info output
        optim - transformers.optimization.AdamW; optimizer
        scheduler - scheduling params
        n_epochs - number of epochs to train

        """ # noqa
        self._model.train()

        if optim is None:
            log.info("No opt. provided, setting all params with AdamW.")
            optim = AdamW(self._model.parameters(), lr=5e-5)
        else:
            log.info("Optimizer provided")

        if scheduler is None:
            log.info("No scheduler provided.")
        else:
            log.info("Scheduler provided.")

        started = time.time()
        for epoch in range(n_epochs):
            total_loss = 0

            for batch in dataset:
                optim.zero_grad()

                with LightwoodAutocast():
                    inpids = batch["input_ids"].to(self.device)
                    attn = batch["attention_mask"].to(self.device)
                    labels = batch["labels"].to(self.device)
                    outputs = self._model(inpids, attention_mask=attn, labels=labels)
                    loss = outputs[0]

                total_loss += loss.item()

                loss.backward()
                optim.step()
                if scheduler is not None:
                    scheduler.step()
                if time.time() - started > self.stop_after:
                    break

            if time.time() - started > self.stop_after:
                break
            self._train_callback(epoch, total_loss / len(dataset))

    def _train_callback(self, epoch, loss):
        log.info(f"{self.name} at epoch {epoch+1} and loss {loss}!")

    def encode(self, column_data: Iterable[str]) -> torch.Tensor:
        """
        Converts each text example in a column into encoded state. This can be either a vector embedding of the [CLS] token (represents the full text input) OR the logits prediction of the output.

        The transformer model is of form:
        transformer base + pre-classifier linear layer + classifier layer

        The embedding returned is of the [CLS] token after the pre-classifier layer; from internal testing, we found the latent space most highly separated across classes. 

        If the encoder represents the logits in classification, returns a soft-maxed output of the class vector.

        :param column_data: List of text data as strings
        :returns: Embedded vector N_rows x Nembed_dim OR logits vector N_rows x N_classes depending on if `embed_mode` is True or not.
        """ # noqa
        if self.is_prepared is False:
            raise Exception("You need to first prepare the encoder.")

        # Set model to testing/eval mode.
        self._model.eval()

        encoded_representation = []

        with torch.no_grad():
            # Set the weights; this is GPT-2
            for text in column_data:

                # Omit NaNs
                if is_none(text):
                    text = ""

                # Tokenize the text with the built-in tokenizer.
                inp = self._tokenizer.encode(
                    text, truncation=True, return_tensors="pt"
                ).to(self.device)

                if self.embed_mode:  # Embedding mode ON; return [CLS]
                    output = self._model.base_model(inp).last_hidden_state[:, 0]

                    # If the model has a pre-classifier layer, use this embedding.
                    if hasattr(self._model, "pre_classifier"):
                        output = self._model.pre_classifier(output)

                else:  # Embedding mode off; return classes
                    output = self._model(inp).logits

                encoded_representation.append(output.detach())

        return torch.stack(encoded_representation).squeeze(1).to('cpu')

    def decode(self, encoded_values_tensor, max_length=100):
        """
        Text generation via decoding is not supported.
        """ # noqa
        raise Exception("Decoder not implemented.")

    def to(self, device, available_devices):
        """
        Converts encoder models to device specified (CPU/GPU)

        Transformers are LARGE models, please run on GPU for fastest implementation.
        """ # noqa
        for v in vars(self):
            attr = getattr(self, v)
            if isinstance(attr, torch.nn.Module):
                attr.to(device)
        return self



import unittest
import numpy as np
import random
import torch
from torch.nn.functional import softmax
from sklearn.metrics import accuracy_score
from lightwood.encoder import BinaryEncoder, NumericEncoder
from lightwood.encoder.text import PretrainedLangEncoder
from lightwood.api.dtype import dtype
import pandas as pd
import os
import pathlib


def create_synthetic_data(n, ptrain=0.7):
    """
    Returns "N" instances of a fake language.

    Labels are 0/1; 0 -> negative sentiment, 1-> positive sentiment
    
    Creates a nonsense string of positive/negative words with some random subset of them.

    :param n: the maximum character (n-1) in the string
    """ # noqa

    textdir = str(pathlib.Path(__file__).parent.resolve())

    # Pick positive/negative words
    with open(os.path.join(textdir, "pos.txt"), "r") as f:
        pos_list = f.readlines()
        pos_list = [i.strip("\n") for i in pos_list]

    with open(os.path.join(textdir, "neg.txt"), "r") as f:
        neg_list = f.readlines()
        neg_list = [i.strip("\n") for i in neg_list]

    data = []
    label = []

    for i in range(n):
        y = random.randint(0, 1)

        # Negative words
        if y == 0:
            word = random.choice(neg_list)
        # Positive words
        else:
            word = random.choice(pos_list)

        data.append(word)
        label.append(y)

    Ntrain = int(n * ptrain)
    train = pd.DataFrame([data[:Ntrain], label[:Ntrain]]).T
    test = pd.DataFrame([data[Ntrain:], label[Ntrain:]]).T

    train.columns = ["text", "label"]
    test.columns = ["text", "label"]
    return train, test


class TestPretrainedLangEncoder(unittest.TestCase):
    def test_encode_and_decode(self):
        """
        Test end-to-end training of values. 

        Uses labeled data with positive/negative sentiment. Model outputs logits, since embed_mode is false. Should be better than random, as transformer language models may actually capture semantics (This is a hypothesis from priming lit.)
        """ # noqa
        seed = 2

        np.random.seed(seed)
        torch.manual_seed(seed)
        random.seed(seed)

        # Make priming data:
        train, test = create_synthetic_data(1000)
        output_enc = BinaryEncoder(is_target=True)
        output_enc.prepare(train["label"])
        encoded_target_values = output_enc.encode(train["label"])

        # Prepare the language encoder
        enc = PretrainedLangEncoder(stop_after=10, embed_mode=False, output_type=dtype.binary)
        enc.prepare(train["text"], pd.DataFrame(), encoded_target_values=encoded_target_values)

        test_labels = test["label"].tolist()
        pred_labels = softmax(enc.encode(test["text"]), dim=1).argmax(dim=1).tolist()

        encoder_accuracy = accuracy_score(test_labels, pred_labels)

        # Should be non-random since models have primed associations to sentiment
        print(f'Categorial encoder accuracy for: {encoder_accuracy} on testing dataset')

    def test_embed_mode(self):
        """
        Test if embed-mode is triggered when flagged.
        Checks if returned embeddings are of size N_rows x N_embed_dim
        """ # noqa
        seed = 2

        np.random.seed(seed)
        torch.manual_seed(seed)
        random.seed(seed)

        # Make priming data:
        train, test = create_synthetic_data(1000)
        output_enc = BinaryEncoder(is_target=True)
        output_enc.prepare(train["label"])
        encoded_target_values = output_enc.encode(train["label"])

        # Prepare the language encoder
        enc = PretrainedLangEncoder(stop_after=10, embed_mode=True, output_type=dtype.binary)
        enc.prepare(train["text"], pd.DataFrame(), encoded_target_values=encoded_target_values)

        # Embeddings of size N_vocab x N_embed_dim for most models (assumes distilbert)
        N_embed_dim = enc._model.base_model.embeddings.word_embeddings.weight.shape[-1]
        embeddings = enc.encode(test["text"])
        assert(embeddings.shape[0] == test.shape[0])
        assert(embeddings.shape[1] == N_embed_dim)

    def test_auto_embed_mode(self):
        """
        For regression (non-categorical type output), text defaults to embed mode as fine-tuning didn't seem to help. Check to see if the output is in fact the size of an embedding. 

        We pretend "embed_mode" is not True, but it should auto-override.

        Expected behavior:
        - does not train the transformer
        - flips transformer to 'embed_mode'
        - returns embedding size N_rows x N_embed_dim
        """ # noqa
        seed = 5

        np.random.seed(seed)
        torch.manual_seed(seed)
        random.seed(seed)

        # Make priming data:
        train, test = create_synthetic_data(2000)
        output_enc = NumericEncoder(is_target=True)
        output_enc.prepare(train["label"])
        encoded_target_values = output_enc.encode(train["label"])

        # Prepare the language encoder
        enc = PretrainedLangEncoder(stop_after=10, embed_mode=False, output_type=dtype.float)
        enc.prepare(train["text"], pd.DataFrame(), encoded_target_values=encoded_target_values)

        # Embeddings of size N_vocab x N_embed_dim for most models (assumes distilbert)
        N_embed_dim = enc._model.base_model.embeddings.word_embeddings.weight.shape[-1]
        embeddings = enc.encode(test["text"])
        assert(enc.embed_mode)
        assert(embeddings.shape[0] == test.shape[0])
        assert(embeddings.shape[1] == N_embed_dim)





import os
import unittest

from torch import Tensor

from lightwood.encoder.audio import MFCCEncoder


class TestMFCCEncoder(unittest.TestCase):
    def test_encode(self):
        if MFCCEncoder is None:
            print('Skipping this test since the system for the encoder work are not installed')
            return

        dir_path = os.path.dirname(os.path.realpath(__file__))
        audio_paths = [
            os.path.join(dir_path, 'test_audio_1.wav'),
            os.path.join(dir_path, 'test_audio_2.wav'),
            None
        ]

        encoder = MFCCEncoder()
        encoder.prepare(audio_paths)
        encoded_audio = encoder.encode(audio_paths)

        self.assertTrue(isinstance(encoded_audio, Tensor))
        # We expect the first dimension to equal the number of images
        self.assertEqual(encoded_audio.size(0), 3)
        self.assertEqual(encoded_audio.size(1), 2000)



# Encoders which should always work
from lightwood.encoder.base import BaseEncoder
from lightwood.encoder.datetime.datetime import DatetimeEncoder
from lightwood.encoder.datetime.datetime_sin_normalizer import DatetimeNormalizerEncoder
from lightwood.encoder.image.img_2_vec import Img2VecEncoder
from lightwood.encoder.numeric.numeric import NumericEncoder
from lightwood.encoder.numeric.ts_numeric import TsNumericEncoder
from lightwood.encoder.array.ts_num_array import TsArrayNumericEncoder
from lightwood.encoder.text.short import ShortTextEncoder
from lightwood.encoder.text.vocab import VocabularyEncoder
from lightwood.encoder.text.rnn import RnnEncoder as TextRnnEncoder
from lightwood.encoder.categorical.onehot import OneHotEncoder
from lightwood.encoder.categorical.binary import BinaryEncoder
from lightwood.encoder.categorical.autoencoder import CategoricalAutoEncoder
from lightwood.encoder.time_series.ts import TimeSeriesEncoder
from lightwood.encoder.array.array import ArrayEncoder, NumArrayEncoder, CatArrayEncoder
from lightwood.encoder.categorical.multihot import MultiHotEncoder
from lightwood.encoder.array.ts_cat_array import TsCatArrayEncoder
from lightwood.encoder.text.pretrained import PretrainedLangEncoder
from lightwood.encoder.audio import MFCCEncoder


__all__ = ['BaseEncoder', 'DatetimeEncoder', 'Img2VecEncoder', 'NumericEncoder', 'TsNumericEncoder',
           'TsArrayNumericEncoder', 'ShortTextEncoder', 'VocabularyEncoder', 'TextRnnEncoder', 'OneHotEncoder',
           'CategoricalAutoEncoder', 'TimeSeriesEncoder', 'ArrayEncoder', 'MultiHotEncoder', 'TsCatArrayEncoder',
           'NumArrayEncoder', 'CatArrayEncoder',
           'PretrainedLangEncoder', 'BinaryEncoder', 'DatetimeNormalizerEncoder', 'MFCCEncoder']



import unittest

import numpy as np
import pandas as pd

from lightwood.data.timeseries_analyzer import get_naive_residuals
from lightwood.helpers.general import mase, evaluate_array_accuracy


class TestTransformTS(unittest.TestCase):
    def test_mase(self):
        true = np.array([[1.0, 2.0, 3.0, 4.0, 5.0]])

        # edge case: perfect forecast
        for scale_error in [1e0, 1e2, 1e4]:
            self.assertTrue(mase(true, true, scale_error, fh=5) == 0)

        # check naive forecast is exactly one
        naive_residual = np.average(abs(true[:, 1:] - true[:, :-1]))
        self.assertTrue(mase(true[:, 1:], true[:, :-1], naive_residual, fh=4) == 1)

        # edge case: constant series
        true = np.array([[2.0, 2.0, 2.0, 2.0, 2.0]])
        pred = np.array([[4.0, 4.0, 4.0, 4.0, 4.0]])
        self.assertTrue(mase(true, pred, 0.0, fh=5) == 2.0)

        # test multiple instance handling (i.e. two 5-step-ahead forecasts)
        true = [[10, 20, 30, 40, 50], [60, 70, 80, 90, 100]]
        pred = [[15, 25, 35, 45, 55], [65, 75, 85, 95, 105]]
        self.assertTrue(mase(true, pred, scale_error=5, fh=5) == 1)
        self.assertTrue(mase(true, pred, scale_error=1, fh=5) == 5)
        self.assertTrue(mase(true, pred, scale_error=10, fh=5) == 0.5)

    def test_get_residuals(self):
        data_len = 10

        target = [i for i in range(data_len)]
        all_residuals, mean = get_naive_residuals(pd.DataFrame(target))
        self.assertEqual(all_residuals, [1.0 for _ in range(data_len - 1)])
        self.assertEqual(mean, 1)

        target = [0 for _ in range(data_len)]
        all_residuals, mean = get_naive_residuals(pd.DataFrame(target))
        self.assertEqual(all_residuals, [0.0 for _ in range(data_len - 1)])
        self.assertEqual(mean, 0)

        target = [1, 4, 2, 5, 3]
        all_residuals, mean = get_naive_residuals(pd.DataFrame(target))
        self.assertEqual(all_residuals, [3.0, 2.0, 3.0, 2.0])
        self.assertEqual(mean, 2.5)

    def test_evaluate_array_r2_accuracy(self):
        true = np.array([[10, 20, 30, 40, 50], [60, 70, 80, 90, 100]])
        self.assertTrue(evaluate_array_accuracy(true, true) == 1.0)

        pred = np.array([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
        self.assertTrue(evaluate_array_accuracy(true, pred) == 0.0)

        pred = np.array([[i + 1 for i in instance] for instance in true])
        self.assertGreaterEqual(evaluate_array_accuracy(true, pred), 0.99)

        pred = np.array([[i - 1 for i in instance] for instance in true])
        self.assertGreaterEqual(evaluate_array_accuracy(true, pred), 0.99)

        pred = np.array([[-i for i in instance] for instance in true])
        self.assertTrue(evaluate_array_accuracy(true, pred) == 0.0)



from copy import deepcopy
from typing import Dict, List, Union

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.lightgbm import LightGBM
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class LightGBMArray(BaseMixer):
    """LightGBM-based model, intended for usage in time series tasks."""
    models: List[LightGBM]
    horizon: int
    submodel_stop_after: float
    target: str
    supports_proba: bool

    def __init__(
            self, stop_after: float, target: str, dtype_dict: Dict[str, str],
            input_cols: List[str],
            horizon: int, fit_on_dev: bool, target_encoder: BaseEncoder):
        super().__init__(stop_after)
        self.submodel_stop_after = stop_after / horizon
        self.target = target
        # dtype_dict[target] = dtype.float  @TODO: figure out if this can be removed
        self.models = [LightGBM(self.submodel_stop_after, target, dtype_dict, input_cols, fit_on_dev,
                                False, target_encoder)
                       for _ in range(horizon)]
        self.horizon = horizon  # for time series tasks, how long is the forecast horizon
        self.supports_proba = False
        self.stable = True

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info('Started fitting LGBM models for array prediction')
        original_target_train = deepcopy(train_data.data_frame[self.target])
        original_target_dev = deepcopy(dev_data.data_frame[self.target])

        for timestep in range(self.horizon):
            if timestep > 0:
                train_data.data_frame[self.target] = train_data.data_frame[f'{self.target}_timestep_{timestep}']
                dev_data.data_frame[self.target] = dev_data.data_frame[f'{self.target}_timestep_{timestep}']

            self.models[timestep].fit(train_data, dev_data)  # @TODO: this call could be parallelized

        # restore target
        train_data.data_frame[self.target] = original_target_train
        dev_data.data_frame[self.target] = original_target_dev

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info('Updating array of LGBM models...')
        original_target_train = deepcopy(train_data.data_frame[self.target])
        original_target_dev = deepcopy(dev_data.data_frame[self.target])

        for timestep in range(self.horizon):
            if timestep > 0:
                train_data.data_frame[self.target] = train_data.data_frame[f'{self.target}_timestep_{timestep}']
                dev_data.data_frame[self.target] = dev_data.data_frame[f'{self.target}_timestep_{timestep}']

            self.models[timestep].partial_fit(train_data, dev_data)  # @TODO: this call could be parallelized

        # restore target
        train_data.data_frame[self.target] = original_target_train
        dev_data.data_frame[self.target] = original_target_dev

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        if args.predict_proba:
            log.warning('This model does not output probability estimates')

        length = sum(ds.encoded_ds_lenghts) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=[f'prediction_{i}' for i in range(self.horizon)])

        for timestep in range(self.horizon):
            ydf[f'prediction_{timestep}'] = self.models[timestep](ds, args)['prediction']

        ydf['prediction'] = ydf.values.tolist()
        return ydf[['prediction']]



from typing import Dict
import pandas as pd
import numpy as np
import datetime
from dateutil.parser import parse as parse_dt
from lightwood.api import StatisticalAnalysis, ProblemDefinition
from lightwood.helpers.numeric import filter_nan_and_none
from lightwood.helpers.seed import seed
from lightwood.data.cleaner import cleaner
from lightwood.helpers.log import log
from lightwood.api.dtype import dtype
from scipy.stats import entropy
from lightwood.data.cleaner import _clean_float


def get_datetime_histogram(data: pd.Series, bins: int) -> Dict[str, list]:
    """Generates the histogram for date and datetime types
    """
    if isinstance(data[0], float) or isinstance(data[0], int):
        data = [_clean_float(x) for x in data]
    else:
        data = [_clean_float(parse_dt(str(x)).timestamp()) for x in data]

    Y, X = np.histogram(data, bins=min(bins, len(set(data))),
                        range=(min(data), max(data)), density=False)

    X = X[:-1].tolist()
    Y = Y.tolist()

    X = [str(datetime.datetime.fromtimestamp(x)) for x in X]
    return {
        'x': X,
        'y': Y
    }


def get_numeric_histogram(data: pd.Series, data_dtype: dtype, bins: int) -> Dict[str, list]:
    """Generate the histogram for integer and float typed data
    """
    # Handle arrays that are actual arrays and not things that become arrays later
    if isinstance(data[0], list):
        new_data = []
        for arr in data:
            new_data.extend(arr)
        data = new_data

    data = [_clean_float(x) for x in data]

    Y, X = np.histogram(data, bins=min(bins, len(set(data))),
                        range=(min(data), max(data)), density=False)
    if data_dtype == dtype.integer:
        Y, X = np.histogram(data, bins=[int(round(x)) for x in X], density=False)

    X = X[:-1].tolist()
    Y = Y.tolist()

    return {
        'x': X,
        'y': Y
    }


def compute_entropy_biased_buckets(histogram):
    S, biased_buckets = None, None
    if histogram is not None or len(histogram['x']) == 0:
        hist_x = histogram['x']
        hist_y = histogram['y']
        nr_values = sum(hist_y)
        S = entropy([x / nr_values for x in hist_y], base=max(2, len(hist_y)))
        if S < 0.25:
            pick_nr = -max(1, int(len(hist_y) / 10))
            biased_buckets = [hist_x[i] for i in np.array(hist_y).argsort()[pick_nr:]]
    return S, biased_buckets


def statistical_analysis(data: pd.DataFrame,
                         dtypes: Dict[str, str],
                         identifiers: Dict[str, object],
                         problem_definition: ProblemDefinition,
                         exceptions: list = ['__mdb_original_index'],
                         seed_nr: int = 420) -> StatisticalAnalysis:
    seed(seed_nr)
    log.info('Starting statistical analysis')
    df = cleaner(data, dtypes, problem_definition.pct_invalid,
                 identifiers, problem_definition.target, 'train', problem_definition.timeseries_settings,
                 problem_definition.anomaly_detection)
    columns = [col for col in df.columns if col not in exceptions]

    missing = {col: len([x for x in df[col] if x is None]) / len(df[col]) for col in columns}
    distinct = {col: len(set([str(x) for x in df[col]])) / len(df[col]) for col in columns}

    nr_rows = len(df)
    target = problem_definition.target
    positive_domain = False
    # get train std, used in analysis
    if dtypes[target] in [dtype.float, dtype.integer, dtype.num_tsarray, dtype.quantity]:
        df_std = df[target].astype(float).std()
        if min(df[target]) >= 0:
            positive_domain = True
    elif dtypes[target] in [dtype.num_array]:
        try:
            all_vals = []
            for x in df[target]:
                all_vals += x
            if min(all_vals) >= 0:
                positive_domain = True
            df_std = pd.Series(all_vals).astype(float).std()
        except Exception as e:
            log.warning(e)
            df_std = 1.0
    else:
        df_std = 1.0

    histograms = {}
    buckets = {}
    # Get histograms for each column
    for col in columns:
        histograms[col] = None
        buckets[col] = None
        if dtypes[col] in (dtype.categorical, dtype.binary, dtype.tags):
            hist = dict(df[col].value_counts())
            histograms[col] = {
                'x': list([str(x) for x in hist.keys()]),
                'y': list(hist.values())
            }
            buckets[col] = histograms[col]['x']
        elif dtypes[col] in (dtype.integer, dtype.float, dtype.num_tsarray, dtype.quantity):
            histograms[col] = get_numeric_histogram(filter_nan_and_none(df[col]), dtypes[col], 50)
            buckets[col] = histograms[col]['x']
        elif dtypes[col] in (dtype.date, dtype.datetime):
            histograms[col] = get_datetime_histogram(filter_nan_and_none(df[col]), 50)
        # @TODO: case for num_ and cat_ arrays
        else:
            histograms[col] = {'x': ['Unknown'], 'y': [len(df[col])]}
            buckets[col] = []

    # get observed classes, used in analysis
    target_class_distribution = None
    target_weights = None
    if dtypes[target] in (dtype.categorical, dtype.binary, dtype.cat_tsarray):
        target_class_distribution = dict(df[target].value_counts().apply(lambda x: x / len(df[target])))
        target_weights = {}
        for k in target_class_distribution:
            target_weights[k] = 1 / target_class_distribution[k]
        train_observed_classes = list(target_class_distribution.keys())
    elif dtypes[target] == dtype.tags:
        train_observed_classes = None  # @TODO: pending call to tags logic -> get all possible tags
    else:
        train_observed_classes = None

    bias = {}
    for col in columns:
        S, biased_buckets = compute_entropy_biased_buckets(histograms[col])
        bias[col] = {
            'entropy': S,
            'description': """Under the assumption of uniformly distributed data (i.e., same probability for Head or Tails on a coin flip) mindsdb tries to detect potential divergences from such case, and it calls this "potential bias". Thus by our data having any potential bias mindsdb means any divergence from all categories having the same probability of being selected.""", # noqa
            'biased_buckets': biased_buckets
        }

    avg_words_per_sentence = {}
    for col in columns:
        if dtypes[col] in (dtype.rich_text, dtype.short_text):
            words_per_sentence = []
            for item in df[col]:
                if item is not None:
                    words_per_sentence.append(len(item.split(' ')))
            avg_words_per_sentence[col] = int(np.mean(words_per_sentence))
        else:
            avg_words_per_sentence[col] = None

    log.info('Finished statistical analysis')
    return StatisticalAnalysis(
        nr_rows=nr_rows,
        df_target_stddev=df_std,
        train_observed_classes=train_observed_classes,
        target_class_distribution=target_class_distribution,
        target_weights=target_weights,
        positive_domain=positive_domain,
        histograms=histograms,
        buckets=buckets,
        missing=missing,
        distinct=distinct,
        bias=bias,
        avg_words_per_sentence=avg_words_per_sentence
    )



import copy
import datetime
import dateutil
import numpy as np
import pandas as pd
import multiprocessing as mp
from lightwood.helpers.parallelism import get_nr_procs
from functools import partial
from typing import Dict
from lightwood.api.types import TimeseriesSettings
from lightwood.helpers.log import log
from lightwood.api import dtype


def transform_timeseries(
        data: pd.DataFrame, dtype_dict: Dict[str, str],
        timeseries_settings: TimeseriesSettings, target: str, mode: str) -> pd.DataFrame:
    """
    Block that transforms the dataframe of a time series task to a convenient format for use in posterior phases like model training.
    
    The main transformations performed by this block are:
      - Type casting (e.g. to numerical for `order_by` columns).
      - Windowing functions for historical context based on `TimeseriesSettings.window` parameter.
      - Explicitly add target columns according to the `TimeseriesSettings.horizon` parameter.
      - Flag all rows that are "predictable" based on all `TimeseriesSettings`.
      - Plus, handle all logic for the streaming use case (where forecasts are only emitted for the last observed data point).
    
    :param data: Dataframe with data to transform.
    :param dtype_dict: Dictionary with the types of each column.
    :param timeseries_settings: A `TimeseriesSettings` object.
    :param target: The name of the target column to forecast.
    :param mode: Either "train" or "predict", depending on what phase is calling this procedure.
    
    :return: A dataframe with all the transformations applied.
    """  # noqa

    tss = timeseries_settings
    original_df = copy.deepcopy(data)
    gb_arr = tss.group_by if tss.group_by is not None else []
    ob_arr = tss.order_by
    window = tss.window

    if tss.use_previous_target and target not in data.columns:
        raise Exception(f"Cannot transform. Missing historical values for target column {target} (`use_previous_target` is set to True).")  # noqa

    for hcol in tss.historical_columns:
        if hcol not in data.columns or data[hcol].isna().any():
            raise Exception(f"Cannot transform. Missing values in historical column {hcol}.")

    if '__mdb_forecast_offset' in original_df.columns:
        """ This special column can be either None or an integer. If this column is passed, then the TS transformation will react to the values within:

        * If all rows = `None`, proceed as usual. This ends up generating one HORIZON-length forecast for each row in the DF.
        * If all rows have the same value `N <= 0`, then cutoff the dataframe latest `-N` rows after TS shaping and prime the DF (with `__make_predictions` column) so that a forecast is generated only for the last row (thus more efficient). This enables `WHERE T = LATEST - K` with `0 <= K < WINDOW` syntax upstream in MindsDB.
        * If all rows have the same value `N = 1`, then activate streaming inference mode where a single forecast will be emitted for the timestamp inferred by the `_ts_infer_next_row` method. This enables the (already supported) `WHERE T > LATEST` syntax.
        """  # noqa
        index = original_df[~original_df['__mdb_forecast_offset'].isin([None])]  # trigger if col is constant & != None
        offset_available = index.shape[0] == len(original_df) and \
            original_df['__mdb_forecast_offset'].unique().tolist() != [None]
        if offset_available:
            offset = min(int(original_df['__mdb_forecast_offset'].unique()[0]), 1)
        else:
            offset = 0
        infer_mode = offset_available and offset == 1
        original_df = original_df.reset_index(drop=True) if infer_mode else original_df
    else:
        offset_available = False
        offset = 0
        infer_mode = False

    original_index_list = []
    idx = 0
    for row in original_df.itertuples():
        if _make_pred(row) or infer_mode:
            original_df.at[row.Index, '__make_predictions'] = True
            original_index_list.append(idx)
            idx += 1
        else:
            original_df.at[row.Index, '__make_predictions'] = False
            original_index_list.append(None)

    original_df['original_index'] = original_index_list

    secondary_type_dict = {}
    for col in ob_arr:
        if dtype_dict[col] in (dtype.date, dtype.integer, dtype.float):
            secondary_type_dict[col] = dtype_dict[col]

    # Convert order_by columns to numbers (note, rows are references to mutable rows in `original_df`)
    for i, row in original_df.iterrows():
        for col in ob_arr:
            # @TODO: Remove if the TS encoder can handle `None`
            if row[col] is None or pd.isna(row[col]):
                original_df.at[i, col] = 0.0
            else:
                if dtype_dict[col] == dtype.date:
                    try:
                        row[col] = dateutil.parser.parse(
                            row[col],
                            **{}
                        )
                    except (TypeError, ValueError):
                        pass

                if isinstance(row[col], datetime.datetime):
                    row[col] = row[col].timestamp()

                try:
                    row[col] = float(row[col])
                except ValueError:
                    raise ValueError(f'Failed to order based on column: "{col}" due to faulty value: {row[col]}')

    for oby in tss.order_by:
        original_df[f'__mdb_original_{oby}'] = original_df[oby]

    group_lengths = []
    if len(gb_arr) > 0:
        df_arr = []
        for _, df in original_df.groupby(gb_arr):
            df_arr.append(df.sort_values(by=ob_arr))
            group_lengths.append(len(df))
    else:
        df_arr = [original_df.sort_values(by=ob_arr)]
        group_lengths.append(len(original_df))

    n_groups = len(df_arr)
    last_index = original_df['original_index'].max()
    for i, subdf in enumerate(df_arr):
        if '__mdb_forecast_offset' in subdf.columns and mode == 'predict':
            if infer_mode:
                df_arr[i] = _ts_infer_next_row(subdf, ob_arr, last_index)
                make_preds = [False for _ in range(max(0, len(df_arr[i]) - 1))] + [True]
                last_index += 1
            elif offset_available:
                # truncate to forecast up until some len(df) + offset (which is <= 0)
                new_index = df_arr[i].index[:len(df_arr[i].index) + offset]
                make_preds = [False for _ in range(max(0, len(new_index) - 1))] + [True]
                df_arr[i] = df_arr[i].loc[new_index]
            else:
                make_preds = [True for _ in range(len(df_arr[i]))]
            df_arr[i]['__make_predictions'] = make_preds

    if len(original_df) > 500:
        # @TODO: restore possibility to override this with args
        nr_procs = get_nr_procs(original_df)
        log.info(f'Using {nr_procs} processes to reshape.')
        pool = mp.Pool(processes=nr_procs)
        # Make type `object` so that dataframe cells can be python lists
        df_arr = pool.map(partial(_ts_to_obj, historical_columns=ob_arr + tss.historical_columns), df_arr)
        df_arr = pool.map(partial(_ts_order_col_to_cell_lists,
                          order_cols=ob_arr + tss.historical_columns), df_arr)
        df_arr = pool.map(
            partial(
                _ts_add_previous_rows, order_cols=ob_arr + tss.historical_columns, window=window),
            df_arr)

        df_arr = pool.map(partial(_ts_add_future_target, target=target, horizon=tss.horizon,
                                  data_dtype=tss.target_type, mode=mode),
                          df_arr)

        if tss.use_previous_target:
            df_arr = pool.map(
                partial(_ts_add_previous_target, target=target, window=tss.window),
                df_arr)
        pool.close()
        pool.join()
    else:
        for i in range(n_groups):
            df_arr[i] = _ts_to_obj(df_arr[i], historical_columns=ob_arr + tss.historical_columns)
            df_arr[i] = _ts_order_col_to_cell_lists(df_arr[i], order_cols=ob_arr + tss.historical_columns)
            df_arr[i] = _ts_add_previous_rows(df_arr[i],
                                              order_cols=ob_arr + tss.historical_columns, window=window)
            df_arr[i] = _ts_add_future_target(df_arr[i], target=target, horizon=tss.horizon,
                                              data_dtype=tss.target_type, mode=mode)
            if tss.use_previous_target:
                df_arr[i] = _ts_add_previous_target(df_arr[i], target=target, window=tss.window)

    combined_df = pd.concat(df_arr)

    if '__mdb_forecast_offset' in combined_df.columns:
        combined_df = pd.DataFrame(combined_df[combined_df['__make_predictions']])  # filters by True only
        del combined_df['__make_predictions']

    if not infer_mode and any([i < tss.window for i in group_lengths]):
        if tss.allow_incomplete_history:
            log.warning("Forecasting with incomplete historical context, predictions might be subpar")
        else:
            raise Exception(f'Not enough historical context to make a timeseries prediction (`allow_incomplete_history` is set to False). Please provide a number of rows greater or equal to the window size - currently (number_rows, window_size) = ({min(group_lengths)}, {tss.window}). If you can\'t get enough rows, consider lowering your window size. If you want to force timeseries predictions lacking historical context please set the `allow_incomplete_history` timeseries setting to `True`, but this might lead to subpar predictions depending on the mixer.') # noqa

    df_gb_map = None
    if n_groups > 1:
        df_gb_list = list(combined_df.groupby(tss.group_by))
        df_gb_map = {}
        for gb, df in df_gb_list:
            df_gb_map['_' + '_'.join(str(gb))] = df

    timeseries_row_mapping = {}
    idx = 0

    if df_gb_map is None:
        for _, row in combined_df.iterrows():
            if not infer_mode:
                timeseries_row_mapping[idx] = int(
                    row['original_index']) if row['original_index'] is not None and not np.isnan(
                    row['original_index']) else None
            else:
                timeseries_row_mapping[idx] = idx
            idx += 1
    else:
        for gb in df_gb_map:
            for _, row in df_gb_map[gb].iterrows():
                if not infer_mode:
                    timeseries_row_mapping[idx] = int(
                        row['original_index']) if row['original_index'] is not None and not np.isnan(
                        row['original_index']) else None
                else:
                    timeseries_row_mapping[idx] = idx

                idx += 1

    del combined_df['original_index']

    return combined_df


def _ts_infer_next_row(df: pd.DataFrame, ob: str, last_index: int) -> pd.DataFrame:
    """
    Adds an inferred next row for streaming mode purposes.

    :param df: dataframe from which next row is inferred.
    :param ob: `order_by` column.
    :param last_index: index number of the latest row in `df`.

    :return: Modified `df` with the inferred row appended to it.
    """
    last_row = df.iloc[[-1]].copy()
    if df.shape[0] > 1:
        butlast_row = df.iloc[[-2]]
        delta = (last_row[ob].values - butlast_row[ob].values).flatten()[0]
    else:
        delta = 1
    last_row.original_index = None
    last_row.index = [last_index + 1]
    last_row['__mdb_ts_inferred'] = True
    last_row[ob] += delta
    return df.append(last_row)


def _make_pred(row) -> bool:
    """
    Indicates whether a prediction should be made for `row` or not.
    """
    return not hasattr(row, '__mdb_forecast_offset') or row.make_predictions


def _ts_to_obj(df: pd.DataFrame, historical_columns: list) -> pd.DataFrame:
    """
    Casts all historical columns in a dataframe to `object` type.

    :param df: Input dataframe
    :param historical_columns: Historical columns to type cast

    :return: Dataframe with `object`-typed historical columns
    """
    for hist_col in historical_columns:
        df.loc[:, hist_col] = df[hist_col].astype(object)
    return df


def _ts_order_col_to_cell_lists(df: pd.DataFrame, order_cols: list) -> pd.DataFrame:
    """
    Casts all data in `order_by` columns into cells.

    :param df: Input dataframe
    :param order_cols: `order_by` columns

    :return: Dataframe with all `order_cols` modified so that their values are cells, e.g. `1` -> `[1]`
    """
    for order_col in order_cols:
        for ii in range(len(df)):
            label = df.index.values[ii]
            df.at[label, order_col] = [df.at[label, order_col]]
    return df


def _ts_add_previous_rows(df: pd.DataFrame, order_cols: list, window: int) -> pd.DataFrame:
    """
    Adds previous rows (as determined by `TimeseriesSettings.window`) into the cells of all `order_by` columns.

    :param df: Input dataframe.
    :param order_cols: `order_by` columns.
    :param window: value of `TimeseriesSettings.window` parameter.
    
    :return: Dataframe with all `order_cols` modified so that their values are now arrays of historical context.
    """  # noqa
    for order_col in order_cols:
        for i in range(len(df)):
            previous_indexes = [*range(max(0, i - window), i)]

            for prev_i in reversed(previous_indexes):
                df.iloc[i][order_col].append(
                    df.iloc[prev_i][order_col][-1]
                )

            # Zero pad
            # @TODO: Remove since RNN encoder can do without (???)
            df.iloc[i][order_col].extend(
                [0] * (1 + window - len(df.iloc[i][order_col]))
            )
            df.iloc[i][order_col].reverse()
    return df


def _ts_add_previous_target(df: pd.DataFrame, target: str, window: int) -> pd.DataFrame:
    """
    Adds previous rows (as determined by `TimeseriesSettings.window`) into the cells of the target column.

    :param df: Input dataframe.
    :param target: target column name.
    :param window: value of `TimeseriesSettings.window` parameter.

    :return: Dataframe with new `__mdb_ts_previous_{target}` column that contains historical target context.
    """  # noqa
    if target not in df:
        return df
    previous_target_values = list(df[target])
    del previous_target_values[-1]
    previous_target_values = [None] + previous_target_values

    previous_target_values_arr = []
    for i in range(len(previous_target_values)):
        prev_vals = previous_target_values[max(i - window, 0):i + 1]
        arr = [None] * (window - len(prev_vals) + 1)
        arr.extend(prev_vals)
        previous_target_values_arr.append(arr)

    df[f'__mdb_ts_previous_{target}'] = previous_target_values_arr
    return df


def _ts_add_future_target(df, target, horizon, data_dtype, mode):
    """
    Adds as many columns to the input dataframe as the forecasting horizon asks for (as determined by `TimeseriesSettings.horizon`).

    :param df: Input dataframe.
    :param target: target column name.
    :param horizon: value of `TimeseriesSettings.horizon` parameter.
    :param data_dtype: dictionary with types of all input columns
    :param mode: either "train" or "predict". `Train` will drop rows with incomplet target info. `Predict` has no effect, for now.

    :return: Dataframe with new `{target}_timestep_{i}'` columns that contains target labels at timestep `i` of a total `TimeseriesSettings.horizon`.
    """  # noqa
    if target not in df:
        return df
    if data_dtype in (dtype.integer, dtype.float, dtype.num_array, dtype.num_tsarray):
        df[target] = df[target].astype(float)

    for timestep_index in range(1, horizon):
        next_target_value_arr = list(df[target])
        for del_index in range(0, min(timestep_index, len(next_target_value_arr))):
            del next_target_value_arr[0]
            next_target_value_arr.append(None)
        col_name = f'{target}_timestep_{timestep_index}'
        df[col_name] = next_target_value_arr
        df[col_name] = df[col_name].fillna(value=np.nan)

    return df



from typing import Dict, Tuple, List

import numpy as np
import pandas as pd

from lightwood.api.types import TimeseriesSettings
from lightwood.api.dtype import dtype
from lightwood.encoder.time_series.helpers.common import generate_target_group_normalizers
from lightwood.helpers.general import get_group_matches


def timeseries_analyzer(data: pd.DataFrame, dtype_dict: Dict[str, str],
                        timeseries_settings: TimeseriesSettings, target: str) -> Dict:
    """
    This module analyzes (pre-processed) time series data and stores a few useful insights used in the rest of Lightwood's pipeline.
    
    :param data: dataframe with time series dataset. 
    :param dtype_dict: dictionary with inferred types for every column.
    :param timeseries_settings: A `TimeseriesSettings` object. For more details, check `lightwood.types.TimeseriesSettings`.
    :param target: name of the target column.
    
    The following things are extracted from each time series inside the dataset:
      - group_combinations: all observed combinations of values for the set of `group_by` columns. The length of this list determines how many time series are in the data.
      - deltas: inferred sampling interval 
      - ts_naive_residuals: Residuals obtained from the data by a naive forecaster that repeats the last-seen value. 
      - ts_naive_mae: Mean residual value obtained from the data by a naive forecaster that repeats the last-seen value.
      - target_normalizers: objects that may normalize the data within any given time series for effective learning. See `lightwood.encoder.time_series.helpers.common` for available choices.
    
    :return: Dictionary with the aforementioned insights and the `TimeseriesSettings` object for future references.
    """  # noqa
    tss = timeseries_settings
    info = {
        'original_type': dtype_dict[target],
        'data': data[target].values
    }
    if tss.group_by is not None:
        info['group_info'] = {gcol: data[gcol] for gcol in tss.group_by}  # group col values
    else:
        info['group_info'] = {}

    # @TODO: maybe normalizers should fit using only the training subsets??
    new_data = generate_target_group_normalizers(info)

    if dtype_dict[target] in (dtype.integer, dtype.float, dtype.num_tsarray):
        naive_forecast_residuals, scale_factor = get_grouped_naive_residuals(info, new_data['group_combinations'])
    else:
        naive_forecast_residuals, scale_factor = {}, {}

    deltas = get_delta(data[tss.order_by],
                       info,
                       new_data['group_combinations'],
                       tss.order_by)

    # detect period
    periods, freqs = detect_period(deltas, tss)

    return {'target_normalizers': new_data['target_normalizers'],
            'deltas': deltas,
            'tss': tss,
            'group_combinations': new_data['group_combinations'],
            'ts_naive_residuals': naive_forecast_residuals,
            'ts_naive_mae': scale_factor,
            'periods': periods,
            'sample_freqs': freqs
            }


def get_delta(df: pd.DataFrame, ts_info: dict, group_combinations: list, order_cols: list) -> Dict[str, Dict]:
    """
    Infer the sampling interval of each time series, by picking the most popular time interval observed in the training data.
    
    :param df: Dataframe with time series data.
    :param ts_info: Dictionary used internally by `timeseries_analyzer`. Contains group-wise series information, among other things.
    :param group_combinations: all tuples with distinct values for `TimeseriesSettings.group_by` columns, defining all available time series.
    :param order_cols: all columns specified in `TimeseriesSettings.order_by`. 
    
    :return:
    Dictionary with group combination tuples as keys. Values are dictionaries with the inferred delta for each series, for each `order_col`.
    """  # noqa
    def _most_popular(sorted_keys: pd.Index):
        idx = 0
        while sorted_keys[idx] == 0:
            if idx == len(sorted_keys) - 1:
                break
            else:
                idx += 1
        return sorted_keys[idx]

    deltas = {"__default": {}}

    # get default delta for all data
    for col in order_cols:
        series = pd.Series([x[-1] for x in df[col]])
        series = series.drop_duplicates()  # by this point df is ordered so duplicate timestamps are either because of non-handled groups or repeated data that, for mode delta estimation, should be ignored  # noqa
        rolling_diff = series.rolling(window=2).apply(lambda x: x.iloc[1] - x.iloc[0])
        delta = _most_popular(rolling_diff.value_counts(ascending=False).keys())  # pick most popular
        deltas["__default"][col] = delta

    # get group-wise deltas (if applicable)
    if ts_info.get('group_info', False):
        original_data = ts_info['data']
        for group in group_combinations:
            if group != "__default":
                for col in order_cols:
                    ts_info['data'] = pd.Series([x[-1] for x in df[col]])
                    _, subset = get_group_matches(ts_info, group)
                    if subset.size > 1:
                        rolling_diff = pd.Series(
                            subset.squeeze()).rolling(
                            window=2).apply(
                            lambda x: x.iloc[1] - x.iloc[0])
                        delta = _most_popular(rolling_diff.value_counts(ascending=False).keys())
                        if group in deltas:
                            deltas[group][col] = delta
                        else:
                            deltas[group] = {col: delta}
        ts_info['data'] = original_data

    return deltas


def get_naive_residuals(target_data: pd.DataFrame, m: int = 1) -> Tuple[List, float]:
    """
    Computes forecasting residuals for the naive method (forecasts for time `t` is the value observed at `t-1`).
    Useful for computing MASE forecasting error.

    Note: method assumes predictions are all for the same group combination. For a dataframe that contains multiple
     series, use `get_grouped_naive_resiudals`.

    :param target_data: observed time series targets
    :param m: season length. the naive forecasts will be the m-th previously seen value for each series

    :return: (list of naive residuals, average residual value)
    """  # noqa
    # @TODO: support categorical series as well
    residuals = target_data.rolling(window=m + 1).apply(lambda x: abs(x.iloc[m] - x.iloc[0]))[m:].values.flatten()
    scale_factor = np.average(residuals)
    return residuals.tolist(), scale_factor


def get_grouped_naive_residuals(info: Dict, group_combinations: List) -> Tuple[Dict, Dict]:
    """
    Wraps `get_naive_residuals` for a dataframe with multiple co-existing time series.
    """  # noqa
    group_residuals = {}
    group_scale_factors = {}
    for group in group_combinations:
        idxs, subset = get_group_matches(info, group)
        residuals, scale_factor = get_naive_residuals(pd.DataFrame(subset))  # @TODO: pass m once we handle seasonality
        group_residuals[group] = residuals
        group_scale_factors[group] = scale_factor
    return group_residuals, group_scale_factors


def detect_period(deltas: dict, tss: TimeseriesSettings) -> (Dict[str, float], Dict[str, str]):
    """
    Helper method that, based on the most popular interval for a time series, determines its seasonal peridiocity (sp).
    This bit of information can be crucial for good modelling with methods like ARIMA.
    
    Supported time intervals are:
        * 'year'
        * 'semestral'
        * 'quarter'
        * 'bimonthly'
        * 'monthly'
        * 'weekly'
        * 'daily'
        * 'hourly'
        * 'minute'
        * 'second'

    Note: all computations assume that the first provided `order_by` column is the one that specifies the sp.

    :param deltas: output of `get_delta`, has the most popular interval for each time series.
    :param tss: timeseries settings.

    :return: for all time series 1) a dictionary with its sp and 2) a dictionary with the detected sampling frequency
    """  # noqa
    interval_to_period = {interval: period for (interval, period) in tss.interval_periods}
    secs_to_interval = {
        'year': 60 * 60 * 24 * 365,
        'semestral': 60 * 60 * 24 * 365 // 2,
        'quarter': 60 * 60 * 24 * 365 // 4,
        'bimonthly': 60 * 60 * 24 * 365 // 6,
        'monthly': 60 * 60 * 24 * 31,
        'weekly': 60 * 60 * 24 * 7,
        'daily': 60 * 60 * 24,
        'hourly': 60 * 60,
        'minute': 60,
        'second': 1
    }

    for tag, period in (('year', 1), ('semestral', 2), ('quarter', 4), ('bimonthly', 6), ('monthly', 12),
                        ('weekly', 4), ('daily', 1), ('hourly', 24), ('minute', 1), ('second', 1)):
        if tag not in interval_to_period.keys():
            interval_to_period[tag] = period

    periods = {}
    freqs = {}
    order_col_idx = 0
    for group in deltas.keys():
        delta = deltas[group][tss.order_by[order_col_idx]]
        diffs = [(tag, abs(delta - secs)) for tag, secs in secs_to_interval.items()]
        min_tag, min_diff = sorted(diffs, key=lambda x: x[1])[0]
        periods[group] = interval_to_period.get(min_tag, 1)
        freqs[group] = min_tag

    return periods, freqs



import unittest
import numpy as np
import torch
from lightwood.encoder.time_series.rnn import TimeSeriesEncoder
from lightwood.encoder.helpers import MinMaxNormalizer, CatNormalizer
import pandas as pd


class TestRnnEncoder(unittest.TestCase):

    def test_minmax_normalizer(self):
        data = [[-100.0, -5.0, 0.0, 5.0, 100.0],
                [-1000.0, -50.0, 0.0, 50.0, 1000.0],
                [-500.0, -405.0, -400.0, -395.0, -300.0],
                [300.0, 395.0, 400.0, 405.0, 500.0],
                [0.0, 1e3, 1e6, 1e9, 1e12]]
        normalizer = MinMaxNormalizer()
        normalizer.prepare(data)
        reconstructed = normalizer.decode(normalizer.encode(data))
        self.assertTrue(np.allclose(data, reconstructed, atol=0.1))

    def test_cat_normalizer(self):
        data = [['a', 'b', 'c'],
                ['c', 'b', 'b'],
                ['a', 'a', None]]
        encoded_target = [[[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]],
                          [[0, 0, 0, 1], [0, 0, 1, 0], [0, 0, 1, 0]],
                          [[0, 1, 0, 0], [0, 1, 0, 0], [1, 0, 0, 0]]]
        normalizer = CatNormalizer()
        normalizer.prepare(data)
        encoded = normalizer.encode(data)
        self.assertTrue(encoded_target, encoded)
        dec = [normalizer.decode(e).squeeze().tolist() for e in encoded]
        self.assertTrue(dec[-1][-1] == normalizer.unk)
        dec[-1][-1] = None
        self.assertTrue(data == dec)

    def test_overfit(self):
        series = [[1, 2, 3, 4, 5, 6],
                  [2, 3, 4, 5, 6, 7],
                  [3, 4, 5, 6, 7, 8],
                  [4, 5, 6, 7, 8, 9]]

        example = ([[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6]],  # query
                   [4, 5, 6, 7])                                    # answer

        data = series * 50
        timesteps = 6
        batch_size = 1

        encoder = TimeSeriesEncoder(stop_after=10)
        encoder.prepare(pd.Series(data), pd.Series(data),
                        feedback_hoop_function=lambda x: print(x), batch_size=batch_size)
        encoded = encoder.encode(data)
        decoded = encoder.decode(encoded, steps=timesteps).tolist()

        equal = 0
        unequal = 0
        self.assertEqual(len(data), len(decoded))
        self.assertEqual(len(data[0]), len(decoded[0]))

        for i in range(len(data)):
            for t in range(timesteps):
                if round(decoded[i][0][t], 0) == round(data[i][0][t], 0):
                    equal += 1
                else:
                    unequal += 1

        print(f'Decoder got {equal} correct and {unequal} incorrect')
        self.assertGreaterEqual(equal * 2, unequal)

        error_margin = 10  # 3
        query, answer = example
        encoded_data, preds = encoder.encode(query, get_next_count=1)
        decoded_data = encoder.decode(encoded_data, steps=3).tolist()

        # check reconstruction
        float_query = [list(map(float, q)) for q in query[0]]
        for qry, dec in zip(float_query, decoded_data[0]):
            for truth, pred in zip(qry, dec):
                self.assertGreater(error_margin, abs(truth - pred))

        # check next value prediction
        preds = torch.reshape(preds, (1, -1)).tolist()[-1]
        for ans, pred in zip(answer, preds):
            self.assertGreater(error_margin, abs(pred - ans))



from lightwood.mixer.unit import Unit
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.neural import Neural
from lightwood.mixer.lightgbm import LightGBM
from lightwood.mixer.lightgbm_array import LightGBMArray
from lightwood.mixer.sktime import SkTime
from lightwood.mixer.prophet import ProphetMixer
from lightwood.mixer.regression import Regression

try:
    from lightwood.mixer.qclassic import QClassic
except Exception:
    QClassic = None

__all__ = ['BaseMixer', 'Neural', 'LightGBM', 'LightGBMArray', 'Unit', 'Regression',
           'SkTime', 'QClassic', 'ProphetMixer']



import unittest

import numpy as np
import pandas as pd

from lightwood.helpers.imputers import NumericalImputer, CategoricalImputer


class TestImputers(unittest.TestCase):
    def _load_df(self, cols_to_cat=[]):
        def _to_cat(x):
            return chr(int(x) + 97) if x == x else x

        df = pd.read_csv('tests/data/hdi.csv')
        for col in cols_to_cat:
            df[col] = df[col].apply(_to_cat).astype(str)

        np.random.seed(0)
        idxs = np.random.randint(0, high=len(df), size=(int(len(df) * 0.2)))  # drop some rows at random
        df.loc[idxs] = np.nan
        df.iloc[0] = np.nan  # force first row to be nan for filled value checks
        return df

    def test_numerical(self):
        target = 'Literacy (%)'
        df = self._load_df()
        for value, expected in zip(('mean', 'median', 'mode', 'zero'),
                                   (df[target].mean(), df[target].median(), df[target].dropna().mode().iloc[0], 0)):
            imp = NumericalImputer(target=target, value=value)
            ndf = imp.impute(df)

            assert df[target].isna().any()
            assert not ndf[target].isna().any()
            assert ndf[target].iloc[0] == expected

    def test_categorical(self):

        target = 'Development Index'
        df = self._load_df(cols_to_cat=[target])
        for value, expected in zip(('mode', ), (df[target].dropna().mode().iloc[0], )):
            imp = CategoricalImputer(target=target, value=value)
            ndf = imp.impute(df)

            assert df[target].isna().any()
            assert not ndf[target].isna().any()
            assert ndf[target].iloc[0] == expected



from lightwood.api.dtype import dtype
import unittest
import pandas as pd
from sklearn.metrics import r2_score
from lightwood.api.types import ProblemDefinition


class TestBasic(unittest.TestCase):
    def test_0_predict_file_flow(self):
        from lightwood.api.high_level import json_ai_from_problem, predictor_from_json_ai

        df = pd.read_csv('tests/data/concrete_strength.csv')[:500]
        # Mess with the names to also test if lightwood can deal /w weird names
        df = df.rename(columns={df.columns[1]: f'\'{df.columns[1]}\''})
        df = df.rename(columns={df.columns[2]: f'\'{df.columns[2]}}}'})
        df = df.rename(columns={df.columns[3]: f'{{{df.columns[3]}\"'})
        target = 'concrete_strength'

        # Make this a quantity
        df[target] = [f'{x}$' for x in df[target]]
        pdef = ProblemDefinition.from_dict({'target': target, 'time_aim': 80})
        jai = json_ai_from_problem(df, pdef)
        jai.analysis_blocks = [{
            "module": "ICP",
            "args": {
                "fixed_significance": None,
                "confidence_normalizer": True,  # explicitly test the ICP normalizer in an integration test
                "positive_domain": "$statistical_analysis.positive_domain",
            }},
            {
                "module": "AccStats",
                "args": {"deps": ["ICP"]}
        },
            {
                "module": "ConfStats",
                "args": {"deps": ["ICP"]}
        }]

        predictor = predictor_from_json_ai(jai)
        predictor.learn(df)

        assert predictor.model_analysis.dtypes[target] == dtype.quantity

        predictions = predictor.predict(df)

        # sanity checks
        self.assertTrue(r2_score([float(x.rstrip('$')) for x in df[target]], predictions['prediction']) > 0.8)
        self.assertTrue(all([0 <= p <= 1 for p in predictions['confidence']]))
        self.assertTrue(all([p['lower'] <= p['prediction'] <= p['upper'] for _, p in predictions.iterrows()]))

        # check customizable ICP fixed confidence param
        fixed_conf = 0.8
        fixed_predictions = predictor.predict(df, {'fixed_confidence': fixed_conf})

        assert all([v == fixed_conf for v in fixed_predictions['confidence'].values])

        # test empty dataframe handling
        with self.assertRaises(Exception) as ctx:
            predictor.predict(pd.DataFrame())
        self.assertTrue('Empty input' in str(ctx.exception))



from copy import deepcopy
from typing import List

import pandas as pd

from lightwood.helpers.log import log


class BaseImputer:
    def __init__(self, target: str, dependencies: List[str] = []):
        """
        Lightwood imputers will fill in missing values in any given column.
        
        The single method to implement, `impute`, is where the logic for filling in missing values has to be specified.
        
        Note that if access to other columns is required, this can be specified with the `dependencies` parameter.

        Also note, by default some Lightwood encoders (e.g. categorical) are able to cope with missing data, so having imputers as a general rule is not required but can be a nice to have.

        :param target: Column that the imputer will modify.
        :param dependencies: Any additional columns (other than the target) that will be needed inside `impute()`.
        """  # noqa
        self.target = target
        self.dependencies = dependencies

    def impute(self, data: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("Please implement your custom imputing logic in a module that inherits this class.")


class NumericalImputer(BaseImputer):
    def __init__(self, target: str, dependencies: List[str] = [], value: str = 'zero', typecast: bool = True):
        """
        Imputer for numerical columns. Supports a handful of different approaches to define the imputation value.
        
        :param value: The value to impute. One of 'mean', 'median', 'zero', 'mode'.
        :param typecast: Used to cast the column into 'float' dtype.
        """  # noqa
        self.value = value
        self.typecast = typecast
        super().__init__(target, dependencies)

    def impute(self, data: pd.DataFrame) -> pd.DataFrame:
        data = deepcopy(data)
        col = self.target

        if data[col].dtype not in (int, float):
            if self.typecast:
                try:
                    data[col] = data[col].astype(float)
                except ValueError:
                    log.warning(f'Numerical imputer failed to cast column {col} to float!')
            else:
                log.warning(f'Numerical imputer used in non-numeric column {col} with dtype {data[col].dtype}!')

        if self.value == 'mean':
            value = data[col].dropna().mean()
        elif self.value == 'median':
            value = data[col].dropna().median()
        elif self.value == 'mode':
            value = data[col].dropna().mode().iloc[0]  # if there's a tie, this chooses the smallest value
        else:
            value = 0.0

        data[col] = data[col].fillna(value=value)

        return data


class CategoricalImputer(BaseImputer):
    def __init__(self, target: str, dependencies: List[str] = [], value: str = 'mode'):
        """
        Imputer for categorical columns.
        
        :param value: Type of imputation. Currently, only `mode` is supported, and replaces missing data with the observed mode.
        """  # noqa
        self.value = value
        super().__init__(target, dependencies)

    def impute(self, data: pd.DataFrame) -> pd.DataFrame:
        data = deepcopy(data)
        col = self.target

        if self.value == 'mode':
            value = data[col].dropna().mode().iloc[0]
            data[col] = data[col].fillna(value=value)

        return data



import unittest
import pandas as pd
from lightwood.api.types import ProblemDefinition
from lightwood import dtype


class TestBasic(unittest.TestCase):
    def test_0_unknown_categories_in_test(self):
        """
        This test is meant to check if the explainer can handle a validation set
        with values that it has never seen as part of the training set.
        In order to achieve this, we generate a dataset with a target that contains 600
        values, out of which 100 are distinct. It's basically certain at least 1 of these
        distinct value will be in train and at least one in validate (more likely 90/10 split)
        So if training and predicting goes well, this validates that explainer works with
        new values in the validation set *and* with values in train that never occur in validation.
        Note: This doesn't check how good the confidence is, it'll likely be very bad, we just care
        that it doesn't crash.
        """  # noqa
        from lightwood.api.high_level import predictor_from_problem

        # The target will be categorical and there will be a bunch of values
        # in all datasets (train/dev/validation) that were not present in the others
        df = pd.DataFrame({
            'target': ['_cat' for _ in range(500)] + [f'{i}cat' for i in range(100)],
            'y': [i for i in range(600)]
        })
        target = 'target'

        predictor = predictor_from_problem(df, ProblemDefinition.from_dict(
                                           {'target': target, 'time_aim': 60, 'unbias_target': True,
                                            "pct_invalid": 20}))
        predictor.learn(df)
        assert predictor.model_analysis.dtypes['target'] == dtype.categorical
        predictor.predict(df)



from typing import Dict, Union

from lightwood.mixer.sktime import SkTime


class ProphetMixer(SkTime):
    def __init__(self,
                 stop_after: float,
                 target: str,
                 dtype_dict: Dict[str, str],
                 horizon: int,
                 ts_analysis: Dict,
                 model_path: str = 'fbprophet.Prophet',
                 auto_size: bool = True,
                 hyperparam_search: bool = False,
                 target_transforms: Dict[str, Union[int, str]] = {}
                 ):
        super().__init__(stop_after, target, dtype_dict, horizon, ts_analysis,
                         model_path, auto_size, hyperparam_search, target_transforms)
        self.stable = False
        self.model_path = model_path
        self.possible_models = [self.model_path]
        self.n_trials = len(self.possible_models)



import math
import torch
import numpy as np
from lightwood.encoder.numeric import NumericEncoder
from lightwood.helpers.general import is_none
from lightwood.helpers.log import log


class TsNumericEncoder(NumericEncoder):
    """
    Variant of vanilla numerical encoder, supports dynamic mean re-scaling
    """
    is_timeseries_encoder: bool = True

    def __init__(self, is_target: bool = False, positive_domain: bool = False, grouped_by=None):
        super(TsNumericEncoder, self).__init__(is_target=is_target, positive_domain=positive_domain)
        # time series normalization params
        self.normalizers = None
        self.group_combinations = None
        self.dependencies = grouped_by
        self.output_size = 2 if is_target else 3

    def encode(self, data, dependency_data={}):
        """
        :param dependency_data: dict with grouped_by column info, to retrieve the correct normalizer for each datum
        """  # noqa
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')
        if not dependency_data:
            dependency_data = {'__default': [None] * len(data)}

        ret = []
        for real, group in zip(data, list(zip(*dependency_data.values()))):
            try:
                real = float(real)
            except Exception:
                try:
                    real = float(real.replace(',', '.'))
                except Exception:
                    real = None
            if self.is_target:
                vector = [0] * 2
                if group is not None and self.normalizers is not None:
                    try:
                        mean = self.normalizers[tuple(group)].abs_mean
                    except KeyError:
                        # novel group-by, we use default normalizer mean
                        mean = self.normalizers['__default'].abs_mean
                else:
                    mean = self._abs_mean

                if not is_none(real):
                    vector[0] = 1 if real < 0 and not self.positive_domain else 0
                    vector[1] = real / mean if mean != 0 else real
                else:
                    pass
                    # This should raise an exception *once* we fix the TsEncoder such that this doesn't get feed `nan`
                    # raise Exception(f'Can\'t encode target value: {real}')

            else:
                vector = [0] * 3
                try:
                    if not is_none(real):
                        vector[0] = 1
                        vector[1] = 1 if real < 0 and not self.positive_domain else 0
                        vector[2] = real / self._abs_mean
                except Exception as e:
                    log.error(f'Can\'t encode input value: {real}, exception: {e}')

            ret.append(vector)

        return torch.Tensor(ret)

    def decode(self, encoded_values, decode_log=None, dependency_data=None):
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if decode_log is None:
            decode_log = self.decode_log

        ret = []
        if not dependency_data:
            dependency_data = {'__default': [None] * len(encoded_values)}
        if isinstance(encoded_values, torch.Tensor):
            encoded_values = encoded_values.tolist()

        for vector, group in zip(encoded_values, list(zip(*dependency_data.values()))):
            if self.is_target:
                if np.isnan(vector[0]) or vector[0] == float('inf') or np.isnan(vector[1]) or vector[1] == float('inf'):
                    log.error(f'Got weird target value to decode: {vector}')
                    real_value = pow(10, 63)
                else:
                    if decode_log:
                        sign = -1 if vector[0] > 0.5 else 1
                        try:
                            real_value = math.exp(vector[1]) * sign
                        except OverflowError:
                            real_value = pow(10, 63) * sign
                    else:
                        if group is not None and self.normalizers is not None:
                            try:
                                mean = self.normalizers[tuple(group)].abs_mean
                            except KeyError:
                                # decode new group with default normalizer
                                mean = self.normalizers['__default'].abs_mean
                        else:
                            mean = self._abs_mean

                        real_value = vector[1] * mean if mean != 0 else vector[1]

                    if self.positive_domain:
                        real_value = abs(real_value)

                    if self._type == 'int':
                        real_value = int(round(real_value, 0))

            else:
                if vector[0] < 0.5:
                    ret.append(None)
                    continue

                real_value = vector[2] * self._abs_mean

                if self._type == 'int':
                    real_value = round(real_value)

            ret.append(real_value)
        return ret



from typing import List, Optional

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.helpers.numeric import is_nan_numeric
from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments, SubmodelData
from lightwood.data.encoded_ds import EncodedDs
from lightwood.helpers.general import evaluate_accuracy


class BestOf(BaseEnsemble):
    """
    This ensemble acts as a mixer selector. 
    After evaluating accuracy for all internal mixers with the validation data, it sets the best mixer as the underlying model.
    """  # noqa
    indexes_by_accuracy: List[float]

    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, accuracy_functions,
                 args: PredictionArguments, ts_analysis: Optional[dict] = None) -> None:
        super().__init__(target, mixers, data)

        score_list = []
        for _, mixer in enumerate(mixers):
            score_dict = evaluate_accuracy(
                data.data_frame,
                mixer(data, args)['prediction'],
                target,
                accuracy_functions,
                ts_analysis=ts_analysis
            )

            avg_score = np.mean(list(score_dict.values()))
            log.info(f'Mixer: {type(mixer).__name__} got accuracy: {avg_score}')
            self.submodel_data.append(SubmodelData(
                name=type(mixer).__name__,
                accuracy=avg_score,
                is_best=False
            ))

            if is_nan_numeric(avg_score):
                avg_score = -pow(2, 63)
                log.warning(f'Change the accuracy of mixer {type(mixer).__name__} to valid value: {avg_score}')

            score_list.append(avg_score)

        self.indexes_by_accuracy = list(reversed(np.array(score_list).argsort()))
        self.supports_proba = self.mixers[self.indexes_by_accuracy[0]].supports_proba
        log.info(f'Picked best mixer: {type(self.mixers[self.indexes_by_accuracy[0]]).__name__}')
        self.submodel_data[self.indexes_by_accuracy[0]].is_best = True

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        if args.all_mixers:
            predictions = {}
            for mixer in self.mixers:
                predictions[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']
            return pd.DataFrame(predictions)
        else:
            for mixer_index in self.indexes_by_accuracy:
                mixer = self.mixers[mixer_index]
                try:
                    return mixer(ds, args=args)
                except Exception as e:
                    if mixer.stable:
                        raise(e)
                    else:
                        log.warning(f'Unstable mixer {type(mixer).__name__} failed with exception: {e}.\
                        Trying next best')



import torch
import numpy as np
import random


def seed(seed_nr: int) -> None:
    torch.manual_seed(seed_nr)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    np.random.seed(seed_nr)
    random.seed(seed_nr)



from lightwood.api.dtype import dtype
'''
def is_allowed(v):
    if v is None:
        return True

    if isinstance(v, bool):
        return True

    try:
        float(v)
        return True
    except:
        pass

    if v in ['True', 'False']:
        return True

    if isinstance(v, str):
        if v.startswith('"') and v.endswith('"'):
            return True
        if v.startswith("'") and v.endswith("'"):
            return True

    # Predictor member
    if v.startswith('self.') and '(' not in v and len(v) < 50:
        return True

    # Allowed variable names
    if v in ['df', 'data', 'encoded_data', 'train_data', 'encoded_train_data', 'test_data']:
        return True

    try:
        cv = dict(v)
        for k in cv:
            ka = is_allowed(k)
            ma = is_allowed(cv[k])
            if not ka or not ma:
                return False
        return True
    except Exception:
        pass

    try:
        cv = list(v)
        for ma in cv:
            ma = is_allowed(m)
            if not ma:
                return False
        return True
    except Exception:
        pass

    raise Exception(f'Possible code injection: {v}')
'''


def is_allowed(v):
    if '(' in str(v):
        return False
    if 'lambda' in str(v):
        return False
    if '__' in str(v):
        return False

    return True


def call(entity: dict) -> str:
    # Special behavior for ensemble
    if 'submodels' in entity['args']:
        del entity['args']['submodels']

    for k, v in entity['args'].items():
        if not str(v).startswith('$'):
            if not is_allowed(v):
                raise Exception(f'Invalid value: {v} for arg {k}')

    args = [f'{k}={v}' for k, v in entity['args'].items() if not str(v).startswith('$')]

    for k, v in entity['args'].items():
        if str(v).startswith('$'):
            v = str(v).replace('$', 'self.')
            args.append(f'{k}={v}')

    args = ','.join(args)
    return f"""{entity['module']}({args})"""


def inline_dict(obj: dict) -> str:
    arr = []
    for k, v in obj.items():
        if str(v) in list(dtype.__dict__.keys()):
            v = f"'{v}'"
        k = k.replace("'", "\\'").replace('"', '\\"')
        arr.append(f"""'{k}': {v}""")

    dict_code = '{\n' + ',\n'.join(arr) + '\n}'
    return dict_code


def align(code: str, indent: int) -> str:
    add_space = ''
    for _ in range(indent):
        add_space += '    '

    code_arr = code.split('\n')
    code = f'\n{add_space}'.join(code_arr)
    return code



"""
2021.07.16

For encoders that already fine-tune on the targets (namely text)
the unity mixer just arg-maxes the output of the encoder.
"""

from typing import List

import torch
import pandas as pd

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.encoder.base import BaseEncoder
from lightwood.data.encoded_ds import EncodedDs
from lightwood.api.types import PredictionArguments


class Unit(BaseMixer):
    def __init__(self, stop_after: float, target_encoder: BaseEncoder):
        super().__init__(stop_after)
        self.target_encoder = target_encoder
        self.supports_proba = False
        self.stable = True

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info("Unit Mixer just borrows from encoder")

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        pass

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        if args.predict_proba:
            # @TODO: depending on the target encoder, this might be enabled
            log.warning('This model does not output probability estimates')

        decoded_predictions: List[object] = []

        for X, _ in ds:
            decoded_prediction = self.target_encoder.decode(torch.unsqueeze(X, 0))
            decoded_predictions.extend(decoded_prediction)

        ydf = pd.DataFrame({"prediction": decoded_predictions})
        return ydf



import numpy as np
import pandas as pd
from lightwood.api.types import TimeseriesSettings


def get_inferred_timestamps(df: pd.DataFrame, col: str, deltas: dict, tss: TimeseriesSettings) -> pd.DataFrame:
    horizon = tss.horizon
    if tss.group_by:
        gby = [f'group_{g}' for g in tss.group_by]

    for (idx, row) in df.iterrows():
        last = [r for r in row[f'order_{col}'] if r == r][-1]  # filter out nans (safeguard; it shouldn't happen anyway)

        if tss.group_by:
            try:
                series_delta = deltas[tuple(row[gby].tolist())][col]
            except KeyError:
                series_delta = deltas['__default'][col]
        else:
            series_delta = deltas['__default'][col]
        timestamps = [last + t * series_delta for t in range(horizon)]

        if tss.horizon == 1:
            timestamps = timestamps[0]  # preserves original input format if horizon == 1

        df[f'order_{col}'].iloc[idx] = timestamps
    return df[f'order_{col}']


def add_tn_conf_bounds(data: pd.DataFrame, tss_args: TimeseriesSettings):
    """
    Add confidence (and bounds if applicable) to t+n predictions, for n>1
    TODO: active research question: how to guarantee 1-e coverage for t+n, n>1
    For now, (conservatively) increases width by the confidence times the log of the time step (and a scaling factor).
    """
    for col in ['confidence', 'lower', 'upper']:
        data[col] = data[col].astype(object)

    for idx, row in data.iterrows():
        error_increase = [row['confidence']] + \
                         [row['confidence'] * np.log(np.e + t / 2)  # offset by e so that y intercept is 1
                          for t in range(1, tss_args.horizon)]
        data['confidence'].iloc[idx] = [row['confidence'] for _ in range(tss_args.horizon)]

        preds = row['prediction']
        width = row['upper'] - row['lower']
        data['lower'].iloc[idx] = [pred - (width / 2) * modifier for pred, modifier in zip(preds, error_increase)]
        data['upper'].iloc[idx] = [pred + (width / 2) * modifier for pred, modifier in zip(preds, error_increase)]

    return data



from lightwood.ensemble.base import BaseEnsemble
from lightwood.ensemble.best_of import BestOf
from lightwood.ensemble.mean_ensemble import MeanEnsemble
from lightwood.ensemble.mode_ensemble import ModeEnsemble
from lightwood.ensemble.weighted_mean_ensemble import WeightedMeanEnsemble

__all__ = ['BaseEnsemble', 'BestOf', 'MeanEnsemble', 'ModeEnsemble', 'WeightedMeanEnsemble']



import random
from datetime import datetime
import unittest
import numpy as np
import pandas as pd
from typing import List
from scipy import signal
from sklearn.metrics import r2_score
from lightwood.api.types import ProblemDefinition
from tests.utils.timing import train_and_check_time_aim
from sktime.forecasting.base import ForecastingHorizon
try:
    from sktime.forecasting.statsforecast import StatsForecastAutoARIMA as AutoARIMA
except ModuleNotFoundError:
    from sktime.forecasting.arima import AutoARIMA

from lightwood.api.high_level import json_ai_from_problem, code_from_json_ai, predictor_from_code, predictor_from_problem  # noqa
from lightwood.mixer.sktime import SkTime

np.random.seed(0)


class TestTimeseries(unittest.TestCase):
    def check_ts_prediction_df(self, df: pd.DataFrame, horizon: int, orders: List[str]):
        for idx, row in df.iterrows():
            lower = [row['lower']] if horizon == 1 else row['lower']
            upper = [row['upper']] if horizon == 1 else row['upper']
            prediction = [row['prediction']] if horizon == 1 else row['prediction']

            assert len(prediction) == horizon

            for oby in orders:
                assert len(row[f'order_{oby}']) == horizon
                assert not any(pd.isna(row[f'order_{oby}']))

            for t in range(horizon):
                assert lower[t] <= prediction[t] <= upper[t]

    def split_arrivals(self, data: pd.DataFrame, grouped: bool) -> (pd.DataFrame, pd.DataFrame):
        train_ratio = 0.8

        if grouped:
            group = 'Country'
            train = pd.DataFrame(columns=data.columns)
            test = pd.DataFrame(columns=data.columns)
            for g in data[group].unique():
                subframe = data[data[group] == g]
                length = subframe.shape[0]
                train = train.append(subframe[:int(length * train_ratio)])
                test = test.append(subframe[int(length * train_ratio):])
        else:
            train = data[:int(data.shape[0] * train_ratio)]
            test = data[int(data.shape[0] * train_ratio):]

        return train, test

    def test_0_time_series_grouped_regression(self):
        """Test grouped numerical predictions, with anomalies and forecast horizon > 1 """
        data = pd.read_csv('tests/data/arrivals.csv')
        train, test = self.split_arrivals(data, grouped=True)
        target = 'Traffic'
        order_by = 'T'
        horizon = 2
        window = 5
        jai = json_ai_from_problem(train,
                                   ProblemDefinition.from_dict({'target': target,
                                                                'time_aim': 30,
                                                                'anomaly_detection': True,
                                                                'timeseries_settings': {
                                                                    'use_previous_target': True,
                                                                    'allow_incomplete_history': True,
                                                                    'group_by': ['Country'],
                                                                    'horizon': horizon,
                                                                    'order_by': [order_by],
                                                                    'period_intervals': (('daily', 7),),
                                                                    'window': window
                                                                }}))
        for i, mixer in enumerate(jai.model['args']['submodels']):
            if mixer["module"] == 'SkTime':
                sktime_mixer_idx = i

        jai.model['args']['submodels'][sktime_mixer_idx] = {
            "module": "SkTime",
            "args": {
                "stop_after": "$problem_definition.seconds_per_mixer",
                "horizon": "$problem_definition.timeseries_settings.horizon",
                "model_path": "'trend.TrendForecaster'",  # use a cheap forecaster
                "hyperparam_search": False,  # disable this as it's expensive and covered in test #3
            },
        }

        code = code_from_json_ai(jai)
        pred = predictor_from_code(code)

        # Test with a short time aim
        train_and_check_time_aim(pred, train)
        preds = pred.predict(test)
        self.check_ts_prediction_df(preds, horizon, [order_by])

        # test allowed incomplete history
        preds = pred.predict(test[:window - 1])
        self.check_ts_prediction_df(preds, horizon, [order_by])

        # test inferring mode, check timestamps are further into the future than test dates
        test['__mdb_forecast_offset'] = 1
        preds = pred.predict(test)
        self.check_ts_prediction_df(preds, horizon, [order_by])

        latest_timestamp = pd.to_datetime(test[order_by]).max().timestamp()
        for idx, row in preds.iterrows():
            for timestamp in row[f'order_{order_by}']:
                assert timestamp > latest_timestamp

        # Check custom ICP params
        test.pop('__mdb_forecast_offset')
        preds = pred.predict(test, {'fixed_confidence': 0.01, 'anomaly_cooldown': 100})
        assert all([all([v == 0.01 for v in f]) for f in preds['confidence'].values])
        assert pred.pred_args.anomaly_cooldown == 100

    def test_1_time_series_regression(self):
        np.random.seed(0)
        data = pd.read_csv('tests/data/arrivals.csv')
        train_df, test_df = self.split_arrivals(data, grouped=False)
        target = 'Traffic'
        order_by = 'T'
        horizon = 2
        window = 5
        pred = predictor_from_problem(data,
                                      ProblemDefinition.from_dict({'target': target,
                                                                   'anomaly_detection': False,
                                                                   'timeseries_settings': {
                                                                       'use_previous_target': False,
                                                                       'allow_incomplete_history': False,
                                                                       'horizon': horizon,
                                                                       'order_by': [order_by],
                                                                       'window': window}
                                                                   }))

        # add a few invalid datetime values to test cleaning procedures
        for idx in list(np.where(np.random.random((len(train_df),)) > 0.98)[0]):
            train_df.at[idx, 'T'] = pd.NaT

        pred.learn(train_df)
        preds = pred.predict(data.sample(frac=1)[0:10])
        self.assertTrue('original_index' in preds.columns)
        self.check_ts_prediction_df(preds, horizon, [order_by])

        # test incomplete history, should not be possible
        self.assertRaises(Exception, pred.predict, test_df[:window - 1])

        # test inferring mode, check timestamps are further into the future than test dates
        test_df['__mdb_forecast_offset'] = 1
        test_df = test_df.sample(frac=1)  # shuffle to test internal ordering logic
        preds = pred.predict(test_df)
        self.check_ts_prediction_df(preds, horizon, [order_by])

        latest_timestamp = pd.to_datetime(test_df[order_by]).max().timestamp()
        for idx, row in preds.iterrows():
            for timestamp in row[f'order_{order_by}']:
                assert timestamp > latest_timestamp

        # test null offset mode
        test_df['__mdb_forecast_offset'] = 0
        preds = pred.predict(test_df)
        self.check_ts_prediction_df(preds, horizon, [order_by])
        assert preds.shape[0] == 1
        last_dt = datetime.utcfromtimestamp(preds[f'order_{order_by}'].values[0][0])
        formatted = str(last_dt.year) + '-' + str(last_dt.month)
        assert formatted == test_df.sort_values(by=order_by).iloc[-1][order_by]

    def test_2_time_series_classification_short_horizon_binary(self):
        df = pd.read_csv('tests/data/arrivals.csv')[:127]
        target = 'Traffic'
        df[target] = df[target] > 100000

        train_idxs = np.random.rand(len(df)) < 0.8
        train = df[train_idxs]
        test = df[~train_idxs]

        predictor = predictor_from_problem(df,
                                           ProblemDefinition.from_dict({'target': target,
                                                                        'time_aim': 80,
                                                                        'anomaly_detection': False,
                                                                        'timeseries_settings': {
                                                                            'order_by': ['T'],
                                                                            'use_previous_target': True,
                                                                            'window': 5
                                                                        },
                                                                        }))

        predictor.learn(train)
        predictor.predict(test)

    def test_3_time_series_classification_long_horizon_binary(self):
        df = pd.read_csv('tests/data/arrivals.csv')[:127]
        target = 'Traffic'
        df[target] = df[target] > 100000

        train_idxs = np.random.rand(len(df)) < 0.8
        train = df[train_idxs]
        test = df[~train_idxs]

        predictor = predictor_from_problem(df,
                                           ProblemDefinition.from_dict({'target': target,
                                                                        'time_aim': 80,
                                                                        'anomaly_detection': False,
                                                                        'timeseries_settings': {
                                                                            'order_by': ['T'],
                                                                            'use_previous_target': True,
                                                                            'window': 5,
                                                                            'horizon': 2
                                                                        },
                                                                        }))

        predictor.learn(train)
        predictor.predict(test)

    def test_4_time_series_classification_long_horizon_multiclass(self):
        df = pd.read_csv('tests/data/arrivals.csv')[:127]  # enforce "Country" to be "No information"
        target = 'Traffic'
        df[target] = df[target].apply(lambda x: chr(65 + int(str(x / 10000)[0])))  # multiclass time series target

        # test array columns as additional input
        df['test_num_array'] = [[random.choice([1, 2, 3, 4]) for __ in range(4)] for _ in range(df.shape[0])]
        df['test_cat_array'] = [[random.choice(['a', 'b', 'c', 'd']) for __ in range(4)] for _ in range(df.shape[0])]

        train_idxs = np.random.rand(len(df)) < 0.8
        train = df[train_idxs]
        test = df[~train_idxs]

        predictor = predictor_from_problem(df,
                                           ProblemDefinition.from_dict({'target': target,
                                                                        'time_aim': 80,
                                                                        'anomaly_detection': False,
                                                                        'timeseries_settings': {
                                                                            'order_by': ['T'],
                                                                            'use_previous_target': True,
                                                                            'window': 5,
                                                                            'horizon': 2
                                                                        },
                                                                        }))

        predictor.learn(train)
        predictor.predict(test)

    def test_5_time_series_sktime_mixer(self):
        """
        Tests `sktime` mixer individually, as it has a special notion of
        timestamps that we need to ensure are being used correctly. In
        particular, given a train-dev-test split, any forecasts coming from a sktime
        mixer should start from the latest observed data in the entire dataset.
        
        This test also compares:
         - correct propagation of offset by K if the special `__mdb_forecast_offset` column is present
         - results against manual use of sktime to ensure equal results.
        """  # noqa

        # synth square wave
        tsteps = 100
        target = 'Value'
        horizon = 20
        t = np.linspace(0, 1, tsteps, endpoint=False)
        ts = [i + f for i, f in enumerate(signal.sawtooth(2 * np.pi * 5 * t, width=0.5))]
        df = pd.DataFrame(columns=['Time', target])
        df['Time'] = t
        df[target] = ts
        df[f'{target}_2x'] = [2 * elt for elt in ts]

        train = df[:int(len(df) * 0.8)]
        test = df[int(len(df) * 0.8):]

        pdef = ProblemDefinition.from_dict({'target': target,
                                            'time_aim': 200,
                                            'timeseries_settings': {
                                                'order_by': ['Time'],
                                                'window': 5,
                                                'horizon': horizon,
                                                'historical_columns': [f'{target}_2x']
                                            }})

        json_ai = json_ai_from_problem(df, problem_definition=pdef)
        json_ai.model['args']['submodels'] = [{
            "module": "SkTime",
            "args": {
                "stop_after": "$problem_definition.seconds_per_mixer",
                "horizon": "$problem_definition.timeseries_settings.horizon",
            }}]

        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)

        # Test with a longer time aim
        train_and_check_time_aim(predictor, train)
        ps = predictor.predict(test)
        assert r2_score(test[target].values, ps['prediction'].iloc[0]) >= 0.95

        # test offset for infer mode
        test['__mdb_forecast_offset'] = 1  # one step after latest (inferred)
        ps1 = predictor.predict(test)
        test['__mdb_forecast_offset'] = 0  # at latest
        ps0 = predictor.predict(test)
        test['__mdb_forecast_offset'] = -1  # one step before latest
        psm1 = predictor.predict(test)
        times_1 = psm1['order_Time'].tolist()[0]
        values_1 = psm1['prediction'].tolist()[0]
        times0 = ps0['order_Time'].tolist()[0]
        values0 = ps0['prediction'].tolist()[0]
        times1 = ps1['order_Time'].tolist()[0]
        values1 = ps1['prediction'].tolist()[0]

        # due to the offset, these intermediate indexes should be equal
        self.assertTrue(times_1[1:] == times0[0:-1])
        self.assertTrue(times0[1:] == times1[0:-1])
        self.assertTrue(values_1[1:] == values0[0:-1])
        self.assertTrue(values0[1:] == values1[0:-1])

        # the rest should be different
        self.assertTrue(times_1 != times0)
        self.assertTrue(times0 != times1)
        self.assertTrue(values_1 != values0)
        self.assertTrue(values0 != values1)

        # test historical columns asserts
        test[f'{target}_2x'].iloc[0] = np.nan
        self.assertRaises(Exception, predictor.predict, test)

        test.pop(f'{target}_2x')
        self.assertRaises(Exception, predictor.predict, test)

        # compare vs sktime manual usage
        if isinstance(predictor.ensemble.mixers[predictor.ensemble.best_index], SkTime):
            forecaster = AutoARIMA()
            fh = ForecastingHorizon([i for i in range(int(tsteps * 0.8))], is_relative=True)
            forecaster.fit(train[target], fh=fh)
            manual_preds = forecaster.predict(fh[1:horizon + 1]).tolist()
            lw_preds = [p[0] for p in ps['prediction']]
            assert np.allclose(manual_preds, lw_preds, atol=1)

    def test_6_irregular_series(self):
        """
        Even though the suggestion is to feed regularly sampled series into predictors, this test can still help us
        catch undesired behavior when the recommendation is not followed.
        """  # noqa

        # synth square wave
        tsteps = 100
        target = 'Value'
        horizon = 20
        # added random noise for irregular sampling
        np.random.seed(0)
        t = np.linspace(0, 1, tsteps, endpoint=False) + np.random.uniform(size=(tsteps,), low=-0.005, high=0.005)
        ts = [i + f for i, f in enumerate(signal.sawtooth(2 * np.pi * 5 * t, width=0.5))]
        df = pd.DataFrame(columns=['Time', target])
        df['Time'] = t
        df[target] = ts
        df[f'{target}_2x'] = [2 * elt for elt in ts]

        train = df[:int(len(df) * 0.8)]
        test = df[int(len(df) * 0.8):]

        pdef = ProblemDefinition.from_dict({'target': target,
                                            'time_aim': 200,
                                            'timeseries_settings': {
                                                'order_by': ['Time'],
                                                'window': 5,
                                                'horizon': horizon,
                                                'historical_columns': [f'{target}_2x']
                                            }})

        json_ai = json_ai_from_problem(df, problem_definition=pdef)
        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)

        train_and_check_time_aim(predictor, train)  # Test with a longer time aim

        test['__mdb_forecast_offset'] = 1
        predictor.predict(test)



import warnings
from types import SimpleNamespace
from typing import Dict, Optional, Tuple

import numpy as np
import pandas as pd
from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.api import dtype
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from lightwood.helpers.log import log
from sklearn.preprocessing import LabelEncoder

import shap


class ShapleyValues(BaseAnalysisBlock):
    """
    Analysis block that estimates column importance with SHAP (SHapley Additive exPlanations), a game theoretic approach
    to explain the ouput of any machine learning model. SHAP assigns each feature an importance value for a particular
    prediction.

    Reference:
        https://shap.readthedocs.io/en/stable/
        https://proceedings.neurips.cc/paper/2017/file/8a20a8621978632d76c43dfd28b67767-Paper.pdf
    """
    label_encoder: LabelEncoder

    def __init__(self, deps: Optional[Tuple] = ...):
        super().__init__(deps=deps)
        self.label_encoder = LabelEncoder()

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        log.info('Preparing to compute feature importance values with SHAP')
        ns = SimpleNamespace(**kwargs)

        output_dtype = ns.dtype_dict[ns.target]
        train_data: EncodedDs = ns.train_data

        if output_dtype in (dtype.integer, dtype.float, dtype.quantity):
            pass
        elif output_dtype in (dtype.binary, dtype.categorical, dtype.tags):
            self.label_encoder.fit(train_data.data_frame[ns.target].values)
        else:
            log.warning(f'ShapleyValues analyzers not supported for type: {output_dtype}')
            return info

        def model(x: np.ndarray) -> np.ndarray:
            assert(isinstance(x, np.ndarray))
            df = pd.DataFrame(data=x, columns=train_data.data_frame.columns)
            ds = EncodedDs(encoders=train_data.encoders, data_frame=df, target=train_data.target)

            decoded_predictions = ns.predictor(ds=ds, args=PredictionArguments())
            if output_dtype in (dtype.integer, dtype.float, dtype.quantity):
                encoded_predictions = decoded_predictions['prediction'].values
            elif output_dtype in (dtype.binary, dtype.categorical, dtype.tags):
                encoded_predictions = self.label_encoder.transform(decoded_predictions['prediction'].values)

            return encoded_predictions

        info['shap_explainer'] = shap.KernelExplainer(model=model, data=train_data.data_frame)

        return info

    def explain(self,
                row_insights: pd.DataFrame,
                global_insights: Dict[str, object],
                **kwargs
                ) -> Tuple[pd.DataFrame, Dict[str, object]]:
        log.info('Computing feature importance values with Kernel SHAP method')
        ns = SimpleNamespace(**kwargs)

        shap_explainer = ns.analysis.get('shap_explainer', None)
        if shap_explainer is None:
            return row_insights, global_insights

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            shap_values = shap_explainer.shap_values(ns.data, silent=True)

        shap_values_df = pd.DataFrame(shap_values).rename(
            mapper=lambda i: f"feature_{i}_impact", axis='columns')

        if kwargs.get('target_dtype', None) in (dtype.binary, dtype.categorical, dtype.tags):
            predictions = self.label_encoder.transform(row_insights['prediction'])
        else:
            predictions = row_insights['prediction']

        base_response = (predictions - shap_values_df.sum(axis='columns')).mean()
        global_insights['base_response'] = base_response

        row_insights = pd.concat([row_insights, shap_values_df], axis='columns')

        return row_insights, global_insights
